import typing
import qw_map
import torch


import qandle.config as config
import qandle.operators as op
import qandle.utils as utils
import qandle.qasm as qasm
from qandle.ansaetze.ansatz import BuiltAnsatz


__all__ = ["StronglyEntanglingLayerBudget"]


class StronglyEntanglingLayerBudget(BuiltAnsatz):
    """
    A strongly entangling layer, inspired by `this paper <https://arxiv.org/abs/1804.00633>`_.
    Consists of a series of single-qubit rotations on each qubit specified in the :code:`qubits` parameter, followed by a series of control gates. Instead of specifying a depth, a :code:`param_budget` is specified, which determines the number of parameters in the circuit. The control gates are inserted every :code:`control_gate_spacing` layers, quickly allowing to change the entanglement structure of the circuit.
    """

    def __init__(
        self,
        num_qubits_total: int,
        qubits: typing.Union[typing.List[int], None] = None,
        param_budget: int = 1,
        control_gate=op.CNOT,
        control_gate_spacing=1,
        rotations=["rz", "ry"],
        remapping: typing.Union[typing.Callable, None] = config.DEFAULT_MAPPING,
    ):
        super().__init__()
        assert control_gate in [op.CNOT, op.CZ], f"Control gate {control_gate} not supported"
        if qubits is None:
            qubits = list(range(num_qubits_total))
        self.qubits = qubits
        self.num_qubits = num_qubits_total
        self.rots = [utils.parse_rot(r) for r in rotations]
        if remapping is None:
            remapping = qw_map.none
        self.remapping = remapping
        self.param_budget = param_budget
        self.control_gate = control_gate
        self.control_gate_spacing = control_gate_spacing
        self.layers_ub = self._allocate_params()
        self.layers = torch.nn.Sequential(
            *[layer.build(num_qubits=num_qubits_total) for layer in self.layers_ub]
        )

    def _allocate_params(self):
        def q_yielder():
            def rot_yielder():
                while True:
                    for r in self.rots:
                        yield r

            rots = rot_yielder()
            current_params = 0
            current_depth = 0
            control_iteration = -1
            while True:
                r = next(rots)
                if current_depth % self.control_gate_spacing == 0 and current_depth > 0:
                    control_iteration = (control_iteration + 1) % len(self.qubits)
                    for ci in range(len(self.qubits)):
                        it = control_iteration % (len(self.qubits) - 1)
                        ti = (ci + 1 + it) % len(self.qubits)
                        ci_q, ti_q = self.qubits[ci], self.qubits[ti]
                        yield self.control_gate(control=ci_q, target=ti_q)

                current_depth += 1
                for q in self.qubits:
                    current_params += 1
                    if current_params > self.param_budget:
                        return
                    yield r(qubit=q, remapping=self.remapping)

        return list(q_yielder())

    def __str__(self) -> str:
        return "SEL"

    def to_qasm(self) -> qasm.QasmRepresentation:
        return [layer.to_qasm() for layer in self.layers_ub]

    def decompose(self) -> typing.List[op.UnbuiltOperator]:
        return self.layers_ub

    def forward(self, x):
        return self.layers(x)



import networkx as nx
import random


class Node:
    def __init__(self, control: int, target: int, original_index: int):
        self.c = control
        self.t = target
        self.original_index = original_index
        self.group: int = -1
        self.__hash = hash(f"{self.c}|{self.t}|{self.original_index}|{random.random()}")

    def __repr__(self) -> str:
        return f"{self.c}|{self.t}"

    def __hash__(self) -> int:
        return self.__hash


def groupnodes(G: nx.DiGraph, max_qubits):
    for sub_g in nx.weakly_connected_components(G):
        roots = [n for n in sub_g if G.in_degree(n) == 0]
        if len(roots) > 1:
            G = groupnodes_max_qubits_per_subcircuit(G, max_qubits)  # FIXME
        else:
            G = groupnodes_max_qubits_per_subcircuit(G, max_qubits)
    return G


def groupnodes_max_qubits_per_subcircuit(sub_g: nx.DiGraph, max_qubits: int) -> nx.DiGraph:
    assert max_qubits >= 2, "max_qubits must be >= 2"

    def get_start():
        """Get node with the lowest in_degree, that has group -1"""
        nodes_by_deg = sorted(sub_g.nodes(), key=lambda n: sub_g.in_degree(n))
        for n in nodes_by_deg:
            if n.group == -1:
                return n
        raise ValueError("No node with group -1 found")

    def highest_group() -> int:
        """Get the highest group number"""
        return max(n.group for n in sub_g.nodes())

    def all_parents_grouped(node: Node) -> bool:
        """Check if all parents of node are in a group"""
        return all(pre.group != -1 for pre in sub_g.predecessors(node))

    def add_to_group(start, used_qubits: set):
        used_qubits.add(start.c)
        used_qubits.add(start.t)
        for n in sub_g[start]:  # for all successors
            if n.group == -1:
                if all_parents_grouped(n):
                    if len(used_qubits.union(set([n.c, n.t]))) <= max_qubits:
                        n.group = start.group
                        add_to_group(n, used_qubits)

    while any(n.group == -1 for n in sub_g.nodes()):
        first = get_start()
        first.group = highest_group() + 1
        add_to_group(first, set((first.c, first.t)))

    return sub_g



import torch
import typing
import copy

import qandle.config as config
import qandle.operators as op
import qandle.qasm as qasm
import qandle.ansaetze.ansatz as ansatz

__all__ = ["TwoLocal"]


class TwoLocal(ansatz.UnbuiltAnsatz):
    """
    The 2-local circuit as described in `the Qiskit docs <https://docs.quantum.ibm.com/api/qiskit/qiskit.circuit.library.TwoLocal>`_. It comprises alternating layers of single qubit rotations and CNOT gates. This implementation uses the :code:`linear` entanglement map.
    """

    def __init__(
        self,
        qubits: typing.Union[typing.List[int], None] = None,
        depth: int = 1,
        remapping: typing.Union[typing.Callable, None] = config.DEFAULT_MAPPING,
        q_params=None,
    ):
        super().__init__()
        self.depth = depth
        self.remapping = remapping
        self.q_params = q_params
        self.qubits = qubits

    def build(self, *args, **kwargs) -> "ansatz.BuiltAnsatz":
        return TwoLocalBuilt(
            num_qubits=kwargs["num_qubits"],
            qubits=self.qubits,
            depth=self.depth,
            remapping=self.remapping,
            q_params=self.q_params,
        )

    def __str__(self) -> str:
        return "TL"

    def to_qasm(self) -> str:
        return [g.to_qasm() for g in self.decompose()]  # type: ignore

    def decompose(self) -> typing.List[op.Operator]:
        if self.qubits is None:
            raise ValueError(
                "It is not specified on which qubits to apply the ansatz. Build the circuit to specify or pass them explicitly as a list"
            )
        res = []
        qp = (
            self.q_params if self.q_params is not None else torch.rand(self.depth, len(self.qubits))
        )
        for d in range(self.depth):
            for wi in range(len(self.qubits)):
                res.append(op.RY(qubit=self.qubits[wi], theta=qp[d, wi]))
            for wi in range(len(self.qubits) - 1):
                res.append(op.CNOT(self.qubits[wi], self.qubits[wi + 1]))
        return res


class TwoLocalBuilt(ansatz.BuiltAnsatz):
    def __init__(
        self,
        num_qubits: int,
        qubits: typing.Union[typing.List[int], None],
        depth: int = 1,
        remapping: typing.Union[typing.Callable, None] = config.DEFAULT_MAPPING,
        q_params=None,
    ):
        super().__init__()
        self.num_qubits = num_qubits
        self.depth = depth
        self.remapping = remapping
        self.qubits = qubits or list(range(num_qubits))
        if q_params is None or q_params.shape != (depth, len(self.qubits)):
            q_params = torch.rand(depth, num_qubits)
        self.register_buffer(
            "cnot_matrices", self._get_cnot_matrix(self.qubits, self.num_qubits), persistent=False
        )
        layers = [
            [op.RY(qubit=w, theta=q_params[d, w]).build(num_qubits) for w in range(num_qubits)]
            for d in range(depth)
        ]
        self.mods = torch.nn.ModuleList([torch.nn.Sequential(*lay) for lay in layers])

    @staticmethod
    def _get_cnot_matrix(qubits: typing.List[int], num_qubits: int):
        cnots = [
            op.BuiltCNOT._calculate_matrix(c=qubits[ci], t=qubits[ci + 1], num_qubits=num_qubits)
            for ci in range(len(qubits) - 1)
        ]
        r = cnots[0]
        for c in cnots[1:]:
            r = torch.matmul(r, c)
        return r

    def forward(self, state: torch.Tensor):
        for d in range(self.depth):
            state = self.mods[d](state)
            state = self.cnot_matrices @ state
        return state

    def __str__(self) -> str:
        return "TL"

    def to_qasm(self) -> qasm.QasmRepresentation:
        outp = []
        for d in range(self.depth):
            outp.extend([g.to_qasm() for g in self.mods[d]])
            for ci in range(len(self.qubits) - 1):
                outp.append(op.CNOT(self.qubits[ci], self.qubits[ci + 1]).to_qasm())
        return outp  # type: ignore

    def decompose(self) -> typing.List[op.UnbuiltOperator]:
        outp = []
        for mod in self.mods:
            for seq in mod:
                outp.append(copy.deepcopy(seq))
            for ci in range(len(self.qubits) - 1):
                outp.append(op.CNOT(self.qubits[ci], self.qubits[ci + 1]))
        return outp



import torch
import typing
import math

import qandle
import einops


class QConv(torch.nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: typing.Union[typing.Tuple[int, int], int] = (3, 3),
        padding: typing.Union[typing.Tuple[int, int], int] = (1, 1),
        qdepth: int = 1,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.qdepth = qdepth
        self.kernel_size = (
            kernel_size if isinstance(kernel_size, tuple) else (kernel_size, kernel_size)
        )
        self.padding = padding if isinstance(padding, tuple) else (padding, padding)
        self.unfold = torch.nn.Unfold(kernel_size=kernel_size, padding=self.padding)
        qubits_for_inp = math.ceil(
            math.log2(self.kernel_size[0] * self.kernel_size[1] * in_channels)
        )
        qubits_for_out = math.ceil(math.log2(out_channels))
        self.qubits = max(qubits_for_inp, qubits_for_out, 1)
        amp = qandle.AmplitudeEmbedding(qubits=list(range(self.qubits)), pad_with=0, name="emb")
        sel = qandle.StronglyEntanglingLayer(qubits=list(range(self.qubits)), depth=self.qdepth)
        mes = qandle.MeasureJointProbability()

        self.qcircuit = qandle.Circuit(
            num_qubits=self.qubits,
            layers=[amp, sel, mes],
        )
        self.ein1 = "batch channel feat -> (batch feat) channel"
        self.ein2 = "(batch h_out w_out) channel -> batch channel h_out w_out"
        self.almost_zero = 0.001

    def _post_process(self, x):
        x = x[:, : self.out_channels, :, :]  # remove padding
        x = x * self.out_channels / 2  # rescale
        return x

    def forward(self, x):
        b, c_in, h_in, w_in = x.shape
        h_out = h_in + 2 * self.padding[0] - self.kernel_size[0] + 1  # output height
        w_out = w_in + 2 * self.padding[1] - self.kernel_size[1] + 1  # output width
        if c_in != self.in_channels:
            raise ValueError(f"Input channels {c_in} does not match in_channels {self.in_channels}")
        x = self.unfold(x)
        x = einops.rearrange(x, self.ein1)
        x = x + self.almost_zero  # avoid zero input
        x = self.qcircuit(emb=x)
        x = einops.rearrange(x, self.ein2, batch=b, h_out=h_out, w_out=w_out)
        x = self._post_process(x)
        return x



import typing
import einops.layers.torch as einl
import torch


def reduce_dot(*args):
    """
    Compute the dot product of a series of matrices.
    """
    if len(args) == 1:
        return args[0]
    matrix = args[0]
    if matrix.dim() == 2:
        matrix = matrix.unsqueeze(0)
    for sub_matrix in args[1:]:
        if sub_matrix.dim() == 2:
            sm = sub_matrix.unsqueeze(0)
        elif sub_matrix.dim() == 3:
            sm = sub_matrix
        else:
            raise ValueError(f"Invalid shape {sub_matrix.shape}")
        matrix = torch.matmul(sm, matrix)
    return matrix


def parse_rot(rot: str):
    from . import operators as op

    rot = rot.lower().replace("r", "")
    if rot == "x":
        return op.RX
    elif rot == "y":
        return op.RY
    elif rot == "z":
        return op.RZ
    else:
        raise ValueError(f"Unknown rotation {rot}")


def do_not_implement(*protected, reason="") -> type:
    class LimitedClass(type):
        def __new__(cls, name, bases, attrs):
            for attribute in attrs:
                if attribute in protected:
                    if reason:
                        raise AttributeError(reason)
                    else:
                        raise AttributeError(f'Overiding of attribute "{attribute}" not allowed')
            return super().__new__(cls, name, bases, attrs)

    return LimitedClass


def get_matrix_transforms(
    num_qubits: int, in_circuit: typing.List[int]
) -> typing.Tuple[einl.Rearrange, einl.Rearrange]:
    """
    Get einops layers for transforming between state and matrix representation of a subcircuit.
    """
    qubits_in_subc = [f"a{i}" for i in in_circuit]
    qubits_not_in_subc = [f"a{i}" for i in range(num_qubits) if i not in in_circuit]
    all_qubits = [f"a{i}" for i in range(num_qubits)]
    state_pattern = f"batch ({' '.join(all_qubits)})"
    matrix_pattern = f"(batch {' '.join(qubits_not_in_subc)}) ({' '.join(qubits_in_subc)})"
    to_matrix_pattern = f"{state_pattern} -> {matrix_pattern}"
    to_state_pattern = f"{matrix_pattern} -> {state_pattern}"
    args = {f"a{i}": 2 for i in range(num_qubits)}
    to_matrix = einl.Rearrange(to_matrix_pattern, **args)
    to_state = einl.Rearrange(to_state_pattern, **args)
    return to_matrix, to_state


def __analyze_barren_plateu(
    circuit, loss_f=lambda x: x[0], num_points=30, other_params=0.1
):  # pragma: no cover
    """
    Analyze, whether a circuit has a barren plateau for a given loss function,
    by checking the gradient of the loss function for num_points parameters (while keeping the rest fixed).
    """
    # sel = qandle.StronglyEntanglingLayer(num_qubits=3, depth=8, rotations=["rz", "ry", "rz"], remapping=None)
    # qc =  qandle.QCircuit(3, layers=[*sel.decompose(), ])
    # grads, losses, params, names = analyze_barren_plateu(qc)
    # fig, ax = plt.subplots(ncols=2, sharey=True, sharex=True, figsize=(15, 7))
    # ax[0].imshow(grads,)
    # ax[1].imshow(losses,)
    # ax[0].set_xticks(np.arange(len(params))[::3], labels=[f"{p.item():.2f}" for p in params[::3]])
    # ax[0].set_yticks(np.arange(len(names)), labels=names)
    # fig.tight_layout()

    import torch

    possible_params = torch.linspace(-torch.pi, torch.pi, num_points, requires_grad=True)
    zerosd = {
        k: torch.ones_like(v, requires_grad=True) + other_params
        for k, v in circuit.state_dict().items()
    }
    grads = torch.zeros(len(zerosd), num_points)
    losses = torch.zeros(len(zerosd), num_points)
    for ipname in range(len(zerosd)):
        pname = list(zerosd.keys())[ipname]
        for thetai, theta in enumerate(possible_params):
            sd = zerosd.copy()
            sd[pname] = theta
            circuit.load_state_dict(sd)
            for p in circuit.parameters():
                p.requires_grad = True
            circuit.zero_grad()
            loss = loss_f(circuit())
            loss.backward()
            grad = list(circuit.parameters())[ipname].grad
            grads[ipname, thetai] = grad.abs().mean()
            losses[ipname, thetai] = loss.detach()
    return grads, losses, possible_params.detach(), list(circuit.state_dict().keys())



from . import qcircuit
from . import config
from . import operators as op
import typing
import einops
import re

__all__ = ["draw"]


def draw(circuit) -> str:
    """
    Draw a circuit diagram. This is just a simple text-based drawing, for more advanced drawing, export to OpenQASM and use more advanced tools, like PennyLane's or Qiskit's drawer. For more options, see :class:`qandle.config`.
    """

    def str_(gate):
        if config.DRAW_SHOW_VALUES:
            return str(gate)
        else:
            """remove all digits, dots, underscores, blanks and brackets"""
            return re.sub(r"[\d\._\s\(\)]", "", str(gate))

    if isinstance(circuit, qcircuit.SplittedCircuit):
        """For CuttedQCircuit, draw the subcircuits separately"""
        return _draw_text_splitted(circuit)
    num_qubits = circuit.num_qubits
    qubits = [f"q{i}" for i in range(num_qubits)]
    for gate in circuit.layers:
        _all_w = set(range(num_qubits))
        try:
            qubits[gate.qubit] += config.DRAW_DASH + str_(gate)
            _all_w.remove(gate.qubit)
            for w in _all_w:
                if not config.DRAW_SHIFT_LEFT:
                    qubits[w] += config.DRAW_DASH * (len(str_(gate)) + 1)
        except AttributeError:
            if isinstance(gate, (op.CNOT, op.BuiltCNOT, op.CZ, op.BuiltCZ)):
                _all_w.remove(gate.c)
                _all_w.remove(gate.t)

                # longest line between control and target (including control/target)
                longest = max(
                    len(qubits[li]) for li in range(min(gate.c, gate.t), max(gate.c, gate.t) + 1)
                )

                if config.DRAW_SHIFT_LEFT:
                    qubits[gate.c] += config.DRAW_DASH * (longest - len(qubits[gate.c]))
                    qubits[gate.t] += config.DRAW_DASH * (longest - len(qubits[gate.t]))

                qubits[gate.c] += config.DRAW_DASH + "⬤"
                if isinstance(gate, (op.CNOT, op.BuiltCNOT)):
                    qubits[gate.t] += config.DRAW_DASH + "⊕"
                elif isinstance(gate, (op.CZ, op.BuiltCZ)):
                    qubits[gate.t] += config.DRAW_DASH + "⬤"
                else:
                    raise ValueError("Unknown gate type")
                for w in range(min(gate.c, gate.t) + 1, max(gate.c, gate.t)):
                    if config.DRAW_SHIFT_LEFT:
                        qubits[w] += config.DRAW_DASH * (longest - len(qubits[w]))
                    if config.DRAW_CROSS_BETWEEN_CNOT:
                        qubits[w] += config.DRAW_DASH + "┼"
                    else:
                        qubits[w] += config.DRAW_DASH * 2
                    _all_w.remove(w)
                for w in _all_w:
                    if not config.DRAW_SHIFT_LEFT:
                        qubits[w] += config.DRAW_DASH * 2
            elif isinstance(gate, einops.layers.torch.Rearrange):
                pass  # ignore einops layers
            else:
                for w in _all_w:
                    qubits[w] += config.DRAW_DASH + str_(gate)

    longest = max([len(w) for w in qubits]) + 1
    for w in range(num_qubits):
        qubits[w] += config.DRAW_DASH * (longest - len(qubits[w]))
    return "\n".join(qubits)


def _draw_text_splitted(circuit) -> str:
    num_qubits = circuit.num_qubits
    placeholder = "SUB{}"
    qubits = [f"q{i}" for i in range(num_qubits)]
    sub_drawings = {}
    for subkey, sub in enumerate(circuit.subcircuits):
        stretch = list(sub.mapping.keys())  # list of qubits the subcircuit is applied to
        _all_w = set(range(num_qubits))
        for w in stretch:
            qubits[w] += "-" + placeholder.format(subkey)
            _all_w.remove(w)
        for w in _all_w:
            qubits[w] += "-" + "-" * len(placeholder.format(subkey))
        sub_drawings[subkey] = (
            placeholder.format(subkey)
            + "\n"
            + draw(qcircuit.UnsplittedCircuit(num_qubits=len(stretch), layers=sub.layers))
        )
    for w in range(num_qubits):
        qubits[w] += "-"
    qubits = "\n".join(qubits)
    formatted_sub_drawings = _format_splitted(sub_drawings, how="column")  # type: ignore
    return qubits + "\nSubcircuits:\n" + formatted_sub_drawings


def _format_splitted(sub_drawing: dict, how=typing.Literal["column", "row"]) -> str:
    """
    returns a formatted string of the subcircuits in sub_drawing
    if how == "column", the subcircuits are drawn next to each other
    if how == "row", the subcircuits are drawn below each other
    """
    if how == "row":
        return "\n".join(sub_drawing.values())
    elif how == "column":
        format_str = ""
        result = ""
        longest_block = max(  # Most lines in a sub_drawing
            [len(subv.split("\n")) for subv in sub_drawing.values()]
        )
        sub_drawing = {
            k: v + "\n" * (longest_block - len(v.split("\n"))) for k, v in sub_drawing.items()
        }
        for subkey in sub_drawing:
            longest_line = max([len(la) for la in sub_drawing[subkey].split("\n")])
            format_str += (
                "{" + str(subkey) + ":<" + str(longest_line + config.DRAW_SPLITTED_PAD) + "}"
            )
        for line in zip(*[s.split("\n") for s in sub_drawing.values()]):
            result += format_str.format(*line) + "\n"
        return result
    else:
        raise ValueError("how must be 'column' or 'row'.")



import dataclasses
import typing


@dataclasses.dataclass
class QasmRepresentation:
    """
    A class representing a single gate in a quantum circuit
    """

    gate_str: str
    gate_value: typing.Union[str, float] = ""
    qubit: typing.Union[str, int] = ""
    qasm3_inputs: str = ""
    qasm3_outputs: str = ""

    def __iter__(self):
        yield self

    def to_string(self) -> str:
        s = self.gate_str
        if self.qasm3_inputs != "":
            s += f"({self.qasm3_inputs})"
        else:
            if self.gate_value != "":
                s += f"({self.gate_value})"
        if self.qubit != "" and self.qubit is not None:
            s += f" q[{self.qubit}]"
        s += ";"
        return s


def convert_to_qasm(circuit, qasm_version: int = 2, include_header: bool = True):
    """
    Convert a circuit to qasm
    """

    num_qubits = circuit.num_qubits

    if qasm_version == 2:
        qasm_header = f"""\
OPENQASM 2.0;
include "qelib1.inc";
qreg q[{num_qubits}];
creg c[{num_qubits}];
"""
    else:
        qasm_header = f"""\
OPENQASM 3.0;
qubit[{num_qubits}] q;
bit[{num_qubits}] c;
"""

    reps = circuit.to_qasm()
    outp = ""
    if include_header:
        outp += qasm_header
    if qasm_version == 3:
        for rep in reps:
            for r in rep:
                if r.qasm3_inputs != "":
                    outp += f"input float {r.qasm3_inputs};\n"
                if r.qasm3_outputs != "":
                    outp += f"output float {r.qasm3_outputs};\n"

    for rep in reps:
        for r in rep:
            outp += r.to_string() + "\n"

    return outp



def check(circuit):
    import openqasm3
    import qandle

    q2 = qandle.convert_to_qasm(circuit, qasm_version=2, include_header=True)
    q3 = qandle.convert_to_qasm(circuit, qasm_version=3, include_header=True)
    assert openqasm3.parser.parse(q2).version == "2.0"
    assert openqasm3.parser.parse(q3).version == "3.0"



import pennylane as qml
import torch
import qandle


def test_measureprob():
    @qml.qnode(device=qml.device("default.qubit.torch", wires=3), interface="torch")
    def pl_circuit(inp):
        qml.QubitStateVector(inp, wires=range(3))
        return [qml.probs(w) for w in range(3)]

    inp = torch.rand(2**3, dtype=torch.cfloat)
    inp = inp / torch.linalg.norm(inp, dim=-1, keepdim=True)
    pl_result = pl_circuit(inp)
    pl_result = torch.tensor([r[0] for r in pl_result], dtype=torch.float)
    qandle_mes = qandle.MeasureProbabilityBuilt(num_qubits=3)
    qandle_result = qandle_mes(inp).to(torch.float)
    assert torch.allclose(pl_result, qandle_result)


def test_measureprob_batched():
    @qml.qnode(device=qml.device("default.qubit.torch", wires=3), interface="torch")
    def pl_circuit(inp):
        qml.QubitStateVector(inp, wires=range(3))
        return [qml.probs(w) for w in range(3)]

    inp = torch.rand(17, 2**3, dtype=torch.cfloat)
    inp = inp / torch.linalg.norm(inp, dim=-1, keepdim=True)
    pl_result = pl_circuit(inp)
    pl_result = torch.tensor([list(r[..., 0]) for r in pl_result], dtype=torch.float).T
    qandle_mes = qandle.MeasureProbabilityBuilt(num_qubits=3)
    qandle_result = qandle_mes(inp).to(torch.float)
    assert torch.allclose(pl_result, qandle_result)


def test_measurejoint_batched():
    @qml.qnode(device=qml.device("default.qubit.torch", wires=3), interface="torch")
    def pl_circuit(inp):
        qml.QubitStateVector(inp, wires=range(3))
        return qml.probs(wires=range(3))

    inp = torch.rand(17, 2**3, dtype=torch.cfloat)
    inp = inp / torch.linalg.norm(inp, dim=-1, keepdim=True)
    pl_result = pl_circuit(inp).to(torch.float)

    qandle_mes = qandle.MeasureJointProbability()
    qandle_result = qandle_mes(inp).to(torch.float)
    assert torch.allclose(pl_result, qandle_result)



import pennylane as qml
import torch
import qandle


def test_reset_unbatched():
    for num_qubits in range(1, 5):
        probs = qandle.MeasureProbabilityBuilt(num_qubits=num_qubits)
        for qubit in range(num_qubits):
            reset = qandle.Reset(qubit=qubit)
            assert isinstance(reset.__str__(), str)
            reset.to_qasm()
            resetb = reset.build(num_qubits=num_qubits)
            assert isinstance(resetb.__str__(), str)
            resetb.to_qasm()
            inp = torch.arange(2**num_qubits).to(torch.cfloat) + 0.5 - 0.1j  # non-zero
            inp = inp / torch.linalg.norm(inp)
            inp.requires_grad = True
            probs_before = probs(inp)
            out = resetb(inp)
            probs_after = probs(out)
            if num_qubits == 1:
                assert torch.allclose(probs_after, torch.tensor(1.0))
            else:
                err = f"qubit: {qubit}, num_qubits: {num_qubits}, probs_before: {probs_before}, probs_after: {probs_after}"
                assert torch.allclose(probs_before[:qubit], probs_after[:qubit]), err
                assert torch.allclose(probs_before[qubit + 1 :], probs_after[qubit + 1 :]), err
                assert torch.allclose(probs_after[qubit], torch.tensor(1.0)), err
            assert out.requires_grad
            assert torch.allclose(torch.norm(out), torch.tensor(1.0))


def test_reset_batched():
    for num_qubits in range(1, 5):
        probs = qandle.MeasureProbabilityBuilt(num_qubits=num_qubits)
        for qubit in range(num_qubits):
            reset = qandle.Reset(qubit=qubit)
            resetb = reset.build(num_qubits=num_qubits)
            inp = torch.rand(7, 2**num_qubits, dtype=torch.cfloat)
            inp = inp / torch.linalg.norm(inp, dim=-1, keepdim=True)
            inp.requires_grad = True
            probs_before = probs(inp)
            out = resetb(inp)
            probs_after = probs(out)
            if num_qubits == 1:
                assert torch.allclose(probs_after, torch.tensor(1.0))
            else:
                err = f"qubit: {qubit}, num_qubits: {num_qubits}, probs_before: {probs_before}, probs_after: {probs_after}"
                assert torch.allclose(probs_before[..., :qubit], probs_after[..., :qubit]), err
                assert torch.allclose(
                    probs_before[..., qubit + 1 :], probs_after[..., qubit + 1 :]
                ), err
                assert torch.allclose(probs_after[..., qubit], torch.tensor(1.0)), err
            assert out.requires_grad
            assert torch.allclose(torch.norm(out, dim=-1), torch.tensor(1.0))


def forward_ground_truth(num_w: int, pl_op, inp):
    def pl_circuit():
        qml.QubitStateVector(inp, wires=range(num_w))
        for w in range(num_w):
            qml.RX(phi=0, wires=w)
        pl_op()
        return qml.state()

    dev = qml.device("default.qubit", wires=num_w)
    qnode = qml.QNode(pl_circuit, dev, interface="torch")
    return qnode().to(torch.cfloat)


def test_custom():
    torch.manual_seed(42)
    u = qandle.operators.U(qubit=1, matrix=torch.tensor([[1, 0], [0, 1]]))
    u_b = u.build(num_qubits=3)
    inp = torch.rand(2**3, dtype=torch.cfloat)
    inp = inp / torch.linalg.norm(inp)
    inp.requires_grad = True
    u_b(inp)
    assert isinstance(u.__str__(), str)
    assert isinstance(u_b.__str__(), str)
    u.to_qasm()
    u_b.to_qasm()


def test_unbatched():
    torch.manual_seed(42)
    v = 0.543321
    errors = []
    for num_w in range(6):
        inp = torch.rand(2**num_w, dtype=torch.cfloat)
        inp = inp / torch.linalg.norm(inp)
        inp.requires_grad = True
        for w in range(num_w):
            for pl_op, own_op in (
                (qml.RX, qandle.operators.RX),
                (qml.RY, qandle.operators.RY),
                (qml.RZ, qandle.operators.RZ),
            ):
                gt = forward_ground_truth(num_w, lambda: pl_op(v, wires=w), inp)
                own_rx_u = own_op(qubit=w, theta=v, remapping=None)
                own_rx = own_rx_u.build(num_qubits=num_w)
                own = own_rx(inp)
                assert torch.allclose(gt, own)
                errors.append((gt - own).abs().sum())
                assert isinstance(own_rx.__str__(), str)
                assert isinstance(own_rx_u.__str__(), str)
                own_rx.to_qasm()
                own_rx_u.to_qasm()
            for w2 in range(num_w):
                if w2 != w:
                    gt = forward_ground_truth(num_w, lambda: qml.CNOT(wires=[w, w2]), inp)
                    own_cnot_u = qandle.operators.CNOT(control=w, target=w2)
                    own_cnot = own_cnot_u.build(num_qubits=num_w)
                    own = own_cnot(inp)
                    assert torch.allclose(gt, own)
                    assert isinstance(own_cnot.__str__(), str)
                    assert isinstance(own_cnot_u.__str__(), str)

                    gt = forward_ground_truth(num_w, lambda: qml.CZ(wires=[w, w2]), inp)
                    own_cz_u = qandle.operators.CZ(control=w, target=w2)
                    own_cz = own_cz_u.build(num_qubits=num_w)
                    own = own_cz(inp)
                    assert torch.allclose(gt, own)
                    assert isinstance(own_cz.__str__(), str)
                    assert isinstance(own_cz_u.__str__(), str)

    info = f"errors max: {max(errors)}, avg: {sum(errors)/len(errors)}"
    assert max(errors) < 1e-5, f"Errors too high, {info}"


def test_reuploading():
    q = 3
    v = torch.tensor(0.123)
    torch.manual_seed(42)
    circuit1 = qandle.Circuit(layers=[qandle.RX(qubit=1, theta=v, remapping=None)], num_qubits=q)
    torch.manual_seed(42)
    circuit2 = qandle.Circuit(
        layers=[qandle.RX(qubit=1, name="reupload", remapping=None)], num_qubits=q
    )
    inp_unbatched = torch.rand(2**q, dtype=torch.cfloat)
    inp_unbatched = inp_unbatched / torch.linalg.norm(inp_unbatched)
    unb_1 = circuit1(inp_unbatched)
    unb_2 = circuit2(inp_unbatched, reupload=v)
    assert torch.allclose(unb_1, unb_2), f"unbatched: {unb_1}, {unb_2}, diff {unb_1-unb_2}"

    inp_batched = torch.rand(7, 2**q, dtype=torch.cfloat)
    inp_batched = inp_batched / torch.linalg.norm(inp_batched, dim=-1, keepdim=True)
    bat_1 = circuit1(inp_batched)
    bat_2 = circuit2(inp_batched, reupload=v)
    assert torch.allclose(bat_1, bat_2), f"batched: {bat_1}, {bat_2}, diff {bat_1-bat_2}"
    v_batched = torch.tensor([v, v, v, v, v, v, v])
    bat_3 = circuit2(inp_batched, reupload=v_batched)
    assert torch.allclose(bat_1, bat_3), f"batched: {bat_1}, {bat_3}, diff {bat_1-bat_3}"
    circuit2(inp_unbatched, reupload=v_batched)


def test_batched():
    torch.manual_seed(42)
    v = -2.3456
    errors = []
    for num_w in range(6):
        inp = torch.rand(13, 2**num_w, dtype=torch.cfloat)
        inp = inp / torch.linalg.norm(inp, dim=-1, keepdim=True)
        inp.requires_grad = True
        for w in range(num_w):
            for pl_op, own_op in (
                (qml.RX, qandle.operators.RX),
                (qml.RY, qandle.operators.RY),
                (qml.RZ, qandle.operators.RZ),
            ):
                gt = forward_ground_truth(num_w, lambda: pl_op(v, wires=w), inp)
                own_rx = own_op(qubit=w, theta=v, remapping=None).build(num_qubits=num_w)
                own = own_rx(inp)
                assert torch.allclose(
                    gt, own
                ), f"num_w: {num_w}, qubit: {w}, pl_op: {pl_op}, gt: {gt}, own: {own}, diff: {gt-own}"
                errors.append((gt - own).abs().sum())
            for w2 in range(num_w):
                if w2 != w:
                    gt = forward_ground_truth(num_w, lambda: qml.CNOT(wires=[w, w2]), inp)
                    own_cnot = qandle.operators.CNOT(control=w, target=w2).build(num_qubits=num_w)
                    own = own_cnot(inp)
                    assert torch.allclose(gt, own)
                    errors.append((gt - own).abs().sum())
                if w2 != w:
                    gt = forward_ground_truth(num_w, lambda: qml.SWAP(wires=[w, w2]), inp)
                    own_swap_ = qandle.operators.SWAP(a=w, b=w2)
                    assert isinstance(own_swap_.__str__(), str)
                    own_swap_.to_qasm()
                    own_swap = own_swap_.build(num_qubits=num_w)
                    assert isinstance(own_swap.__str__(), str)
                    own_swap.to_qasm()
                    own = own_swap(inp)
                    assert torch.allclose(gt, own)
                    errors.append((gt - own).abs().sum())
                if w2 != w:
                    gt = forward_ground_truth(num_w, lambda: qml.CZ(wires=[w, w2]), inp)
                    own_cz = qandle.operators.CZ(control=w, target=w2).build(num_qubits=num_w)
                    own = own_cz(inp)
                    assert torch.allclose(gt, own)
                    errors.append((gt - own).abs().sum())
    info = f"errors max: {max(errors)}, avg: {sum(errors)/len(errors)}"
    assert max(errors) < 1e-5, f"Errors too high, {info}"



import qw_map

DEFAULT_MAPPING = qw_map.tanh
r"""
Default weight remapping function for quantum gates. This function limits the weights of the variational quantum circuit's gates. 
:code:`qw_map.tanh` is the default value, yielding better results in practice. For higher performance or naive usage, set :code:`DEFAULT_MAPPING = lambda x: x` or :code:`DEFAULT_MAPPING = qw_map.none`. For more information, see the `arxiv paper 1 <https://arxiv.org/abs/2212.14807>`_ and `arxiv paper 2 <https://arxiv.org/abs/2306.05776>`_.
"""

DRAW_SPLITTED_PAD = 7
"""Number of spaces between columns in the drawer for splitted circuits."""

DRAW_DASH = "─"
r"""
Character used for drawing qubits, can be changed to the normal dash '-', depending on font, where :code:`───` often looks better than :code:`---`

.. code-block:: python

    >>> qandle.config.DRAW_DASH = "─"
    >>> print(circuit.draw())
    q0─RZ─RY─RZ─■─────X─RZ─RY─RZ─■───X───
    q1─RZ─RY─RZ─X─■───┼─RZ─RY─RZ─┼─■─┼─X─
    q2─RZ─RY─RZ───X─■─┼─RZ─RY─RZ─X─┼─■─┼─
    q3─RZ─RY─RZ─────X─■─RZ─RY─RZ───X───■─
    
    >>> qandle.config.DRAW_DASH = "-"
    >>> print(circuit.draw())
    q0-RZ-RY-RZ-■-----X-RZ-RY-RZ-■---X---
    q1-RZ-RY-RZ-X-■---┼-RZ-RY-RZ-┼-■-┼-X-
    q2-RZ-RY-RZ---X-■-┼-RZ-RY-RZ-X-┼-■-┼-
    q3-RZ-RY-RZ-----X-■-RZ-RY-RZ---X---■-

    
"""

DRAW_SHOW_VALUES = True
"""Whether to show the values of the gates in the drawer. If not, only the gate positions are shown, reducing clutter."""

DRAW_SHIFT_LEFT = False
"""
Whether to shift the gates to the left if possible for a more compact drawing.

.. code-block:: python

    >>> qandle.config.DRAW_SHIFT_LEFT = False
    >>> print(circuit.draw())
    q0─RZ─RY─RZ───────────────────■───X─RZ─RY─RZ───────────────────■─X───
    q1──────────RZ─RY─RZ──────────X─■─┼──────────RZ─RY─RZ──────────┼─■─X─
    q2───────────────────RZ─RY─RZ───X─■───────────────────RZ─RY─RZ─X───■─

    >>> qandle.config.DRAW_SHIFT_LEFT = True
    >>> print(circuit.draw())
    q0─RZ─RY─RZ─■───X─RZ─RY─RZ─■─X───
    q1─RZ─RY─RZ─X─■─┼─RZ─RY─RZ─┼─■─X─
    q2─RZ─RY─RZ───X─■─RZ─RY─RZ─X───■─


"""

DRAW_CROSS_BETWEEN_CNOT = True
"""
Whether to draw a cross :code:`┼` on the qubits between the CNOT control and target

.. code-block:: python

    >>> qandle.config.DRAW_CROSS_BETWEEN_CNOT = True
    >>> print(circuit.draw())
    q0─RZ─RY─RZ─■─────X─RZ─RY─RZ─■───X───
    q1─RZ─RY─RZ─X─■───┼─RZ─RY─RZ─┼─■─┼─X─
    q2─RZ─RY─RZ───X─■─┼─RZ─RY─RZ─X─┼─■─┼─
    q3─RZ─RY─RZ─────X─■─RZ─RY─RZ───X───■─

    >>> qandle.config.DRAW_CROSS_BETWEEN_CNOT = False
    >>> print(circuit.draw())
    q0─RZ─RY─RZ─■─────X─RZ─RY─RZ─■───X───
    q1─RZ─RY─RZ─X─■─────RZ─RY─RZ───■───X─
    q2─RZ─RY─RZ───X─■───RZ─RY─RZ─X───■───
    q3─RZ─RY─RZ─────X─■─RZ─RY─RZ───X───■─


"""



# ruff: noqa: F403 F401

from .qcircuit import *
from .measurements import *
from .embeddings import *
from .ansaetze import *
from .drawer import *
from .splitter import *
from . import config
from .convolution import *
from .errors import *
from .operators import *
from .utils import *
from .qasm import *


def __reimport():  # pragma: no cover
    print("reimporting qandle")
    import importlib
    import sys

    modules = {k: v for k, v in sys.modules.items()}
    for module in modules:
        if module.startswith("qandle"):
            importlib.reload(sys.modules[module])

    try:
        # Patch snakeviz to not show in notebook (always open in new tab)
        import snakeviz

        snakeviz.ipymagic._check_ipynb = lambda: False
    except ImportError:
        pass


def __count_parameters(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)



import torch
import abc
import typing
import qw_map
import warnings

import qandle.utils as utils
import qandle.config as config
import qandle.errors as errors
import qandle.qasm as qasm

__all__ = [
    "Operator",
    "RX",
    "RY",
    "RZ",
    "CNOT",
    "CZ",
    "Reset",
    "SWAP",
    "U",
    "CustomGate",
    "Controlled",
    "Invert",
    "BUILT_CLASS_RELATION",
]

matrixbuilder = typing.Tuple[torch.Tensor, torch.Tensor, typing.Callable, typing.Callable]


class AbstractNoForward(abc.ABCMeta, utils.do_not_implement("forward", "backward")):
    pass


class Operator(abc.ABC):
    """Everything that can be applied to a state."""

    named = False

    @abc.abstractmethod
    def __str__(self) -> str:
        """Returns a string representation of the operator."""

    def __repr__(self) -> str:
        return self.__str__()

    @abc.abstractmethod
    def to_qasm(self) -> qasm.QasmRepresentation:
        """Returns the OpenQASM2 representation of the operator."""


class UnbuiltOperator(Operator, abc.ABC):
    """Container class for operators that have not been built yet."""

    @abc.abstractmethod
    def build(self, num_qubits, **kwargs) -> "BuiltOperator":
        """Builds the operator, i.e. converts it to a torch.nn.Module."""


class BuiltOperator(Operator, torch.nn.Module, abc.ABC):
    """Container class for operators that have been built."""

    @abc.abstractmethod
    def forward(self, state: torch.Tensor) -> torch.Tensor:
        """Applies the operator to the state."""

    @abc.abstractmethod
    def to_matrix(self, **kwargs) -> torch.Tensor:
        """Returns the matrix representation of the operator, such that :code:`state @ matrix` is equivalent to :code:`forward(state)`. Might be significantly slower than forward."""


class U(UnbuiltOperator):
    def __init__(self, qubit: int, matrix: torch.Tensor):
        self.qubit = qubit
        self.matrix = matrix

    def __str__(self) -> str:
        return f"U_{self.qubit}"

    def to_qasm(self) -> qasm.QasmRepresentation:
        return qasm.QasmRepresentation(gate_str="U", qubit=self.qubit)

    def build(self, num_qubits, **kwargs) -> "BuiltU":
        return BuiltU(qubit=self.qubit, matrix=self.matrix, num_qubits=num_qubits)


class BuiltU(BuiltOperator):
    def __init__(
        self,
        qubit: int,
        matrix: torch.Tensor,
        num_qubits: int,
        self_description: str = "U",
    ):
        super().__init__()
        self.qubit = qubit
        self.num_qubits = num_qubits
        self.description = self_description

        self.original_matrix = matrix
        m = torch.eye(1)
        for i in range(self.num_qubits):
            m = torch.kron(m, matrix if i == self.qubit else torch.eye(2))
        self.matrix = m.to(torch.cfloat).contiguous()

    def __str__(self) -> str:
        return f"{self.description}_{self.qubit}"

    def to_qasm(self) -> qasm.QasmRepresentation:
        u = f"{self.original_matrix[0, 0]:.2f}, {self.original_matrix[0, 1]:.2f}, {self.original_matrix[1, 0]:.2f}, {self.original_matrix[1, 1]:.2f}"
        definition = (
            "gate "
            + self.description
            + "_"
            + str(self.qubit)
            + " q["
            + str(self.qubit)
            + "] {{ "
            + "U("
            + u
            + ") }}"
        )
        return qasm.QasmRepresentation(gate_str=definition, qubit=self.qubit)

    def forward(self, state: torch.Tensor) -> torch.Tensor:
        return state @ self.matrix

    def to_matrix(self, **kwargs) -> torch.Tensor:
        return self.matrix


CustomGate = BuiltU
"""
Define a custom single qubit gate. 
This module is always built, and therefore might have a big memory footprint. 

Attributes:
    qubit : int
        The index of the qubit this gate operates on.
    matrix : torch.Tensor
        The matrix representation of the gate. Must be a 2x2 unitary matrix.
    num_qubits : int
        The number of qubits in the circuit.
    self_description : str
        The description of the gate. Default is "U". This name will be used in the string representation of the gate and in the OpenQASM2 representation.

"""


class UnbuiltParametrizedOperator(UnbuiltOperator, metaclass=AbstractNoForward):
    """Container class for parametrized operators that have not been built yet."""

    def __init__(
        self,
        qubit: int,
        theta: typing.Union[float, torch.Tensor, None] = None,
        name: typing.Union[str, None] = None,
        **kwargs,
    ):
        """
        Creates a parametrized operator.

        """
        assert isinstance(qubit, int), "qubit must be an integer"
        assert qubit >= 0, "qubit must be >= 0"
        if isinstance(theta, (float, int)):
            theta = torch.tensor(theta, requires_grad=True, dtype=torch.float)
        remapping = kwargs.get("remapping", config.DEFAULT_MAPPING)
        if remapping is None:
            remapping = qw_map.none
        self.qubit = qubit
        self.name = name
        self.named = name is not None
        self.theta = theta
        self.remapping = remapping

    def __str__(self) -> str:
        if self.name is None:  # named
            return f"{self.__class__.__name__}{self.qubit} ({self.name})"
        else:  # unnamed
            if self.theta is None:
                return f"{self.__class__.__name__}{self.qubit}"
            else:
                return f"{self.__class__.__name__}{self.qubit} ({self.remapping(self.theta).item():.2f})"

    def to_qasm(self) -> qasm.QasmRepresentation:
        if self.named:
            return qasm.QasmRepresentation(
                gate_str=self.__class__.__name__.lower(),
                qubit=self.qubit,
                qasm3_inputs=self.name,  # type: ignore # if name is None, named would be False
            )

        if self.theta is not None:
            return qasm.QasmRepresentation(
                gate_str=self.__class__.__name__.lower(),
                qubit=self.qubit,
                gate_value=self.remapping(self.theta).item(),
            )

        raise errors.UnbuiltGateError(
            "This gate has no parameter. Set parameter or build or set name before converting to OpenQASM2."
        )

    def build(self, num_qubits, **kwargs) -> "BuiltParametrizedOperator":
        return BUILT_CLASS_RELATION[self.__class__](
            qubit=self.qubit,
            initialtheta=self.theta,
            name=self.name,
            remapping=self.remapping,
            num_qubits=num_qubits,
        )


class BuiltParametrizedOperator(BuiltOperator, abc.ABC):
    def __init__(
        self,
        qubit: int,
        remapping: typing.Callable,
        num_qubits: int,
        initialtheta: typing.Union[torch.Tensor, None] = None,
        name: typing.Union[str, None] = None,
    ):
        super().__init__()
        self.qubit = qubit
        if initialtheta is None:
            initialtheta = torch.rand(1)
        self.name = name  # currently unused
        self.named = name is not None  # faster than checking if name is None every time
        self.theta = torch.nn.Parameter(initialtheta, requires_grad=True)
        self.remapping = remapping
        self.num_qubits = num_qubits
        self.unbuilt_class = BUILT_CLASS_RELATION.T[self.__class__]

        self.register_buffer("_i", torch.tensor(1j, dtype=torch.cfloat), persistent=False)
        a, b, self.a_op, self.b_op = self.matrix_builder()
        self.register_buffer("_a", a.T.contiguous(), persistent=False)
        self.register_buffer("_b", b.T.contiguous(), persistent=False)

    def __str__(self) -> str:
        base = f"{self.unbuilt_class.__name__}_{self.qubit}"
        if self.named:
            return f"{base} ({self.name})"
        else:
            if self.theta is None:
                return f"{base}"
            else:
                return f"{base} ({self.remapping(self.theta).item():.2f})"

    def to_qasm(self) -> qasm.QasmRepresentation:
        if self.named:
            return qasm.QasmRepresentation(
                gate_str=self.unbuilt_class.__name__.lower(),
                qubit=self.qubit,
                qasm3_inputs=self.name,  # type: ignore # if name is None, named would be False
            )

        else:
            return qasm.QasmRepresentation(
                gate_str=self.unbuilt_class.__name__.lower(),
                qubit=self.qubit,
                gate_value=self.remapping(self.theta).item(),
            )

    def get_matrix(self, **kwargs) -> torch.Tensor:
        if self.named:
            t = kwargs[self.name] / 2  # type: ignore # if name is None, named would be False
            if t.dim() == 1:
                t = t.unsqueeze(-1).unsqueeze(-1)
        else:
            t = self.remapping(self.theta) / 2
        a_matrix = self._a * self.a_op(t)
        b_matrix = self._b * self.b_op(t)
        matrix = a_matrix + b_matrix
        return matrix

    def hydrated(
        self, special: typing.Union[torch.Tensor, None, typing.Tuple] = None
    ) -> torch.Tensor:
        if special is None:
            special = torch.eye(2)
        if not isinstance(special, torch.Tensor):
            special = torch.tensor(special)
        matrix = torch.eye(1)
        for i in range(self.num_qubits):
            matrix = torch.kron(matrix, special if i == self.qubit else torch.eye(2))
        return matrix.to(torch.cfloat)

    def forward(self, state: torch.Tensor, **kwargs) -> torch.Tensor:
        mat = self.get_matrix(**kwargs)
        if mat.dim() == 2:
            res = state @ mat
        else:
            if state.dim() == 1:
                state = state.unsqueeze(0)
                # raise ValueError("One of the named matrices received batched input, but the state is not batched. Please batch the state and the named parameters or neither.")
            res = (state.unsqueeze(1) @ mat).squeeze(1)
        return res

    @abc.abstractmethod
    def matrix_builder(self) -> matrixbuilder:
        """Returns the matrix builder for the gate."""

    def to_matrix(self, **kwargs) -> torch.Tensor:
        return self.get_matrix(**kwargs)


class RX(UnbuiltParametrizedOperator):
    """
    Parametrized RX gate, i.e. a rotation around the x-axis of the Bloch sphere.

    This class represents a parametrized RX gate in a quantum circuit.

    Used by :class:`qandle.qcircuit.QCircuit` to build the circuit.

    Attributes:
        qubit : int
            The index of the qubit this gate operates on.
        theta : float, torch.Tensor, optional
            The parameter of the RX gate, by default None. If None, a random parameter :math:`[0, 1]` is chosen.
        name : str, optional
            The name of the operator, by default None. If None, the operator is not named and does not accept named inputs.
        remapping : Callable, optional
            A function that remaps the parameter theta, by default config.DEFAULT_MAPPING. To disable remapping, pass :code:`qw_map.none` or :code:`lambda x: x`.
    """


class RY(UnbuiltParametrizedOperator):
    """
    Parametrized RY gate, i.e. a rotation around the y-axis of the Bloch sphere.

    This class represents a parametrized RY gate in a quantum circuit.

    Used by :class:`qandle.qcircuit.QCircuit` to build the circuit.

    Attributes
        qubit : int
            The index of the qubit this gate operates on.
        theta : float, torch.Tensor, optional
            The parameter of the RY gate, by default None. If None, a random parameter :math:`[0, 1]` is chosen.
        name : str, optional
            The name of the operator, by default None. If None, the operator is not named and does not accept named inputs.
        remapping : Callable, optional
            A function that remaps the parameter theta, by default config.DEFAULT_MAPPING. To disable remapping, pass :code:`qw_map.none` or :code:`lambda x: x`.
    """


class RZ(UnbuiltParametrizedOperator):
    """
    Parametrized RZ gate, i.e. a rotation around the z-axis of the Bloch sphere.

    This class represents a parametrized RZ gate in a quantum circuit.

    Used by :class:`qandle.qcircuit.QCircuit` to build the circuit.

    Attributes
        qubit : int
            The index of the qubit this gate operates on.
        theta : float, torch.Tensor, optional
            The parameter of the RZ gate, by default None. If None, a random parameter :math:`[0, 1]` is chosen.
        name : str, optional
            The name of the operator, by default None. If None, the operator is not named and does not accept named inputs.
        remapping : Callable, optional
            A function that remaps the parameter theta, by default config.DEFAULT_MAPPING. To disable remapping, pass :code:`qw_map.none` or :code:`lambda x: x`.
    """


class BuiltRX(BuiltParametrizedOperator):
    def matrix_builder(self) -> matrixbuilder:
        return (
            -self.hydrated(((0, 1), (1, 0))) * self._i,
            self.hydrated(),
            torch.sin,
            torch.cos,
        )


class BuiltRY(BuiltParametrizedOperator):
    def matrix_builder(self) -> matrixbuilder:
        return (
            self.hydrated(((0, -1), (1, 0))),
            self.hydrated(),
            torch.sin,
            torch.cos,
        )


class BuiltRZ(BuiltParametrizedOperator):
    def matrix_builder(self) -> matrixbuilder:
        return (
            self.hydrated(((1, 0), (0, 0))),
            self.hydrated(((0, 0), (0, 1))),
            lambda theta: torch.exp(-self._i * theta),
            lambda theta: torch.exp(self._i * theta),
        )


class CNOT(UnbuiltOperator):
    """CNOT gate."""

    def __init__(self, control: int, target: int):
        assert control != target, "Control and target must be different"
        self.c = control
        self.t = target

    def __str__(self) -> str:
        return f"CNOT {self.c}|{self.t}"

    def to_qasm(self) -> qasm.QasmRepresentation:
        return qasm.QasmRepresentation(gate_str=f"cx q[{self.c}], q[{self.t}]")

    def build(self, num_qubits, **kwargs) -> "BuiltCNOT":
        return BuiltCNOT(control=self.c, target=self.t, num_qubits=num_qubits)


class Invert(UnbuiltOperator):
    """
    Special class for inverted gates. Apply the inverse of the target operator.
    """

    def __init__(self, target: Operator):
        self.t = target

    def __str__(self) -> str:
        return f"{self.t}^-1"

    def to_qasm(self) -> qasm.QasmRepresentation:
        return qasm.QasmRepresentation(gate_str=f"{self.t}^-1")

    def build(self, num_qubits, **kwargs) -> "BuiltInvert":
        return BuiltInvert(target=self.t, num_qubits=num_qubits)


class BuiltInvert(BuiltOperator):
    def __init__(self, target: Operator, num_qubits: int):
        super().__init__()
        if hasattr(target, "build"):
            warnings.warn(
                "Building target operator in Invert. If the target operator is parameterized, this might lead to different parameter initialization. Please build the target operator before inverting it."
            )
            target = target.build(num_qubits)
        self.target = target
        self.num_qubits = num_qubits

    def forward(self, state: torch.Tensor) -> torch.Tensor:
        target_matrix = self.target.to_matrix()
        return state @ torch.linalg.inv(target_matrix)

    def __str__(self) -> str:
        return Invert(self.target).__str__()

    def to_qasm(self) -> qasm.QasmRepresentation:
        return Invert(self.target).to_qasm()

    def to_matrix(self, **kwargs) -> torch.Tensor:
        return torch.linalg.inv(self.target.to_matrix())


class Controlled(UnbuiltOperator):
    """
    Special class for controlled gates. Use a control qubit, and apply the target operator if the control qubit is 1.
    """

    def __init__(self, control: int, target: Operator):
        self.c = control
        self.t = target

    def __str__(self) -> str:
        return f"Controlled {self.c}|{self.t}"

    def to_qasm(self) -> qasm.QasmRepresentation:
        return qasm.QasmRepresentation(gate_str=f"controlled q[{self.c}], {self.t}")

    def build(self, num_qubits, **kwargs) -> "BuiltControlled":
        return BuiltControlled(control=self.c, target=self.t, num_qubits=num_qubits)


class BuiltControlled(BuiltOperator):
    def __init__(self, control: int, target: Operator, num_qubits: int):
        super().__init__()
        self.c = control
        if hasattr(target, "build"):
            target = target.build(num_qubits)
        self.t = target
        self.named = target.named
        self.num_qubits = num_qubits

    def forward(self, state: torch.Tensor, **kwargs) -> torch.Tensor:
        target_matrix = self.t.to_matrix(**kwargs)
        c2 = self.num_qubits - self.c - 1
        mask = 1 << c2
        dim = 2**self.num_qubits
        device = state.device
        indices = torch.arange(dim, device=device)
        c1_mask = (indices & mask) != 0
        batched = state.dim() == 1
        if batched:
            state = state.unsqueeze(0)
        state_c1 = state.clone()
        state_c1[:, ~c1_mask] = 0
        # Apply the target operation on the control=1 subspace
        transformed_c1 = torch.squeeze(state_c1.unsqueeze(1) @ target_matrix, 1)
        out = torch.where(c1_mask, transformed_c1, state)
        if batched:
            out = out.squeeze(0)
        return out

    def __str__(self) -> str:
        return Controlled(self.c, self.t).__str__()

    def to_qasm(self) -> qasm.QasmRepresentation:
        return Controlled(self.c, self.t).to_qasm()

    def to_matrix(self, **kwargs) -> torch.Tensor:
        return self.t.to_matrix()


class CZ(UnbuiltOperator):
    """CZ gate."""

    def __init__(self, control: int, target: int):
        assert control != target, "Control and target must be different"
        self.c = control
        self.t = target

    def __str__(self) -> str:
        return f"CZ {self.c}|{self.t}"

    def to_qasm(self) -> qasm.QasmRepresentation:
        return qasm.QasmRepresentation(gate_str=f"cz q[{self.c}], q[{self.t}]")

    def build(self, num_qubits, **kwargs) -> "BuiltCZ":
        return BuiltCZ(control=self.c, target=self.t, num_qubits=num_qubits)


class BuiltCNOT(BuiltOperator):
    def __init__(self, control: int, target: int, num_qubits: int):
        super().__init__()
        self.c = control
        self.t = target
        self.num_qubits = num_qubits
        self.register_buffer(
            "_M", self._calculate_matrix(control, target, num_qubits), persistent=False
        )

    @staticmethod
    def _calculate_matrix(c: int, t: int, num_qubits: int):
        M = torch.zeros(2**num_qubits, 2**num_qubits) * 0j
        c2, t2 = num_qubits - c - 1, num_qubits - t - 1
        for i in range(2**num_qubits):
            if i & (1 << c2):
                M[i, i ^ (1 << t2)] = 1
            else:
                M[i, i] = 1
        return M

    def forward(self, state: torch.Tensor) -> torch.Tensor:
        return state @ self._M

    def __str__(self) -> str:
        return CNOT(self.c, self.t).__str__()

    def to_qasm(self) -> qasm.QasmRepresentation:
        return CNOT(self.c, self.t).to_qasm()

    def to_matrix(self, **kwargs) -> torch.Tensor:
        return self._M


class BuiltCZ(BuiltOperator):
    def __init__(self, control: int, target: int, num_qubits: int):
        super().__init__()
        self.c = control
        self.t = target
        self.num_qubits = num_qubits
        self.register_buffer(
            "_M", self._calculate_matrix(control, target, num_qubits), persistent=False
        )

    @staticmethod
    def _calculate_matrix(c: int, t: int, num_qubits: int):
        c2, t2 = num_qubits - c - 1, num_qubits - t - 1
        indices = torch.arange(2**num_qubits)
        diag = torch.ones(2**num_qubits, dtype=torch.cfloat)
        mask = ((indices & (1 << c2)) != 0) & ((indices & (1 << t2)) != 0)
        diag[mask] = -1
        M = torch.diag(diag)
        return M

    def forward(self, state: torch.Tensor) -> torch.Tensor:
        return state @ self._M

    def __str__(self) -> str:
        return CZ(self.c, self.t).__str__()

    def to_qasm(self) -> qasm.QasmRepresentation:
        return CZ(self.c, self.t).to_qasm()

    def to_matrix(self, **kwargs) -> torch.Tensor:
        return self._M


class BuiltReset(BuiltOperator):
    def __init__(self, qubit: int, num_qubits: int):
        super().__init__()
        self.qubit = qubit
        self.num_qubits = num_qubits
        self.to_matrix_transform, self.to_state_transform = utils.get_matrix_transforms(
            num_qubits, [qubit]
        )

    def forward(self, state: torch.Tensor):
        unbatched = state.dim() == 1
        if unbatched:
            state = state.unsqueeze(0)
        state = self.to_matrix_transform(state)
        old_norm = torch.linalg.norm(state, dim=-1)
        old_norm_0 = (state[:, 0].abs()) + 1e-5
        scale = old_norm / old_norm_0
        new_state = torch.zeros_like(state, dtype=torch.cfloat)
        new_state[:, 0] = state[:, 0] * scale
        state = self.to_state_transform(new_state)
        state = state / torch.linalg.norm(state, dim=-1, keepdim=True)
        if unbatched:
            state = state.squeeze(0)
        return state

    def __str__(self) -> str:
        return Reset(self.qubit).__str__()

    def to_qasm(self) -> qasm.QasmRepresentation:
        return Reset(self.qubit).to_qasm()

    def to_matrix(self, **kwargs) -> torch.Tensor:
        raise ValueError("Reset gate does not have a matrix representation.")


class Reset(UnbuiltOperator):
    """
    Reset gate. Resets the selected qubit to the |0> state, while preserving the norm of the state.
    """

    def __init__(self, qubit: int):
        self.qubit = qubit

    def __str__(self) -> str:
        return f"Reset {self.qubit}"

    def to_qasm(self) -> qasm.QasmRepresentation:
        return qasm.QasmRepresentation(gate_str="reset", qubit=self.qubit)

    def build(self, num_qubits, **kwargs) -> "BuiltReset":
        return BUILT_CLASS_RELATION[self.__class__](qubit=self.qubit, num_qubits=num_qubits)


class BuiltSWAP(BuiltOperator):
    def __init__(self, a: int, b: int, num_qubits: int):
        super().__init__()
        self.a = a
        self.b = b
        self.num_qubits = num_qubits
        self.register_buffer("_M", self._calculate_matrix(a, b, num_qubits), persistent=False)

    @staticmethod
    def _calculate_matrix(a: int, b: int, num_qubits: int):
        swap_matrix = torch.eye(2**num_qubits)
        a, b = num_qubits - a - 1, num_qubits - b - 1
        for i in range(2**num_qubits):
            swapped_i = i
            if ((i >> a) & 1) != ((i >> b) & 1):
                swapped_i = i ^ ((1 << a) | (1 << b))
            swap_matrix[i, i] = 0
            swap_matrix[i, swapped_i] = 1
        return swap_matrix.to(torch.cfloat)

    def forward(self, state: torch.Tensor) -> torch.Tensor:
        return state @ self._M

    def __str__(self) -> str:
        return f"SWAP {self.a}|{self.b}"

    def to_qasm(self) -> qasm.QasmRepresentation:
        return qasm.QasmRepresentation(gate_str=f"swap q[{self.a}], q[{self.b}]")

    def to_matrix(self, **kwargs) -> torch.Tensor:
        return self._M


class SWAP(UnbuiltOperator):
    def __init__(self, a: int, b: int):
        self.a = a
        self.b = b

    def __str__(self) -> str:
        return f"Swap {self.a}|{self.b}"

    def to_qasm(self) -> qasm.QasmRepresentation:
        return qasm.QasmRepresentation(gate_str=f"swap q[{self.a}], q[{self.b}]")

    def build(self, num_qubits, **kwargs) -> "BuiltSWAP":
        return BuiltSWAP(a=self.a, b=self.b, num_qubits=num_qubits)


class rdict(dict):
    """reversible dict"""

    @property
    def T(self):
        return {v: k for k, v in self.items()}


BUILT_CLASS_RELATION = rdict(
    {
        UnbuiltOperator: BuiltOperator,
        UnbuiltParametrizedOperator: BuiltParametrizedOperator,
        RX: BuiltRX,
        RY: BuiltRY,
        RZ: BuiltRZ,
        CNOT: BuiltCNOT,
        Reset: BuiltReset,
        U: BuiltU,
        SWAP: BuiltSWAP,
        CZ: BuiltCZ,
        Invert: BuiltInvert,
        Controlled: BuiltControlled,
    }
)



import torch
from qandle import embeddings
import pennylane as qml


def test_amplitude_embedding_unpadded():
    w = 4
    inp = torch.arange(2**w).to(torch.float)
    emb = embeddings.AmplitudeEmbedding(
        qubits=range(w), normalize=True, pad_with=None, name="amp"
    ).build(num_qubits=w)
    out = emb(amp=inp)
    assert torch.allclose(out.real, inp / inp.norm())
    assert torch.allclose(out.imag, torch.zeros_like(inp))

    @qml.qnode(device=qml.device("default.qubit.torch", wires=w), interface="torch")
    def pl_circuit():
        qml.AmplitudeEmbedding(inp, wires=range(w), normalize=True)
        return qml.state()

    pl_out = pl_circuit().to(torch.cfloat)
    assert torch.allclose(out, pl_out)


def test_amplitude_embedding_unpadded_batched():
    w = 4
    inp = torch.rand(10, 2**w).to(torch.float)
    inp = inp / inp.norm(dim=1, keepdim=True)
    emb = embeddings.AmplitudeEmbedding(
        qubits=range(w), normalize=False, pad_with=None, name="amp"
    ).build(num_qubits=w)
    out = emb(amp=inp)
    assert torch.allclose(out.real, inp)
    assert torch.allclose(out.imag, torch.zeros_like(inp))

    @qml.qnode(device=qml.device("default.qubit.torch", wires=w), interface="torch")
    def pl_circuit():
        qml.AmplitudeEmbedding(inp, wires=range(w), normalize=True)
        return qml.state()

    pl_out = pl_circuit().to(torch.cfloat)
    assert torch.allclose(out, pl_out)


def test_amplitude_embedding_padded():
    inp = torch.rand(11).to(torch.float)
    n_inp = inp / inp.norm()
    emb = embeddings.AmplitudeEmbedding(
        qubits=range(4), normalize=True, pad_with=0, name="amp"
    ).build(num_qubits=4)
    out = emb(amp=inp)
    assert torch.allclose(out.real[:11], n_inp)
    assert torch.allclose(out.imag, torch.zeros(16))


def test_angle_embedding_x3():
    w = 3
    inp = torch.arange(w).to(torch.float) + 0.5

    @qml.qnode(device=qml.device("default.qubit.torch", wires=w), interface="torch")
    def pl_circuit():
        qml.AngleEmbedding(inp, wires=range(w), rotation="X")
        return qml.state()

    pl_out = pl_circuit().to(torch.cfloat)
    emb = embeddings.AngleEmbedding(qubits=list(range(w)), rotation="rx", name="amp").build(
        num_qubits=w
    )
    out = emb(amp=inp)
    assert torch.allclose(out, pl_out)


def test_angle_embedding_x3_batched():
    w = 3
    inp = torch.rand(10, w).to(torch.float) + 0.5

    @qml.qnode(device=qml.device("default.qubit.torch", wires=w), interface="torch")
    def pl_circuit():
        qml.AngleEmbedding(inp, wires=range(w), rotation="X")
        return qml.state()

    pl_out = pl_circuit().to(torch.cfloat)
    emb = embeddings.AngleEmbedding(qubits=list(range(w)), rotation="rx", name="amp").build(
        num_qubits=w
    )
    out = emb(amp=inp)
    assert torch.allclose(out, pl_out)


def test_angle_embedding_y4():
    w = 4
    inp = torch.arange(w).to(torch.float) + 2.5

    @qml.qnode(device=qml.device("default.qubit.torch", wires=w), interface="torch")
    def pl_circuit():
        qml.AngleEmbedding(inp, wires=range(w), rotation="Y")
        return qml.state()

    pl_out = pl_circuit().to(torch.cfloat)
    emb = embeddings.AngleEmbedding(qubits=list(range(w)), rotation="y", name="emb").build(w)
    out = emb(emb=inp)
    assert torch.allclose(out, pl_out)


def test_angle_embedding_y4_batched():
    w = 4
    inp = torch.rand(10, w).to(torch.float) + 2.5

    @qml.qnode(device=qml.device("default.qubit.torch", wires=w), interface="torch")
    def pl_circuit():
        qml.AngleEmbedding(inp, wires=range(w), rotation="Y")
        return qml.state()

    pl_out = pl_circuit().to(torch.cfloat)
    emb = embeddings.AngleEmbedding(qubits=list(range(w)), rotation="y", name="amp").build(
        num_qubits=w
    )
    out = emb(amp=inp)
    assert torch.allclose(out, pl_out)


def test_angle_embedding_z5():
    w = 5
    inp = torch.arange(w).to(torch.float) + 2.5

    @qml.qnode(device=qml.device("default.qubit.torch", wires=w), interface="torch")
    def pl_circuit():
        qml.AngleEmbedding(inp, wires=range(w), rotation="Z")
        return qml.state()

    pl_out = pl_circuit().to(torch.cfloat)
    emb = embeddings.AngleEmbedding(qubits=list(range(w)), rotation="z", name="amp").build(
        num_qubits=w
    )
    out = emb(amp=inp)
    assert torch.allclose(out, pl_out)


def test_angle_embedding_z5_batched():
    w = 5
    inp = torch.rand(10, w).to(torch.float) + 2.5

    @qml.qnode(device=qml.device("default.qubit.torch", wires=w), interface="torch")
    def pl_circuit():
        qml.AngleEmbedding(inp, wires=range(w), rotation="Z")
        return qml.state()

    pl_out = pl_circuit().to(torch.cfloat)
    emb = embeddings.AngleEmbedding(qubits=list(range(w)), rotation="z", name="amp").build(
        num_qubits=w
    )
    out = emb(amp=inp)
    assert torch.allclose(out, pl_out)



import torch
import abc
import typing
import random

import qandle.operators as op
import qandle.utils as utils
import qandle.qasm as qasm


class InputOperator(op.UnbuiltOperator, abc.ABC): ...


class InputOperatorBuilt(op.BuiltOperator, abc.ABC):
    @abc.abstractmethod
    def to_qasm(self) -> qasm.QasmRepresentation: ...

    def decompose(self):
        raise NotImplementedError(f"Decomposing {self.__class__} is not yet supported")

    def to_matrix(self):
        raise ValueError("Input operators do not have a matrix representation")


def _identity(input, *args, **kwargs):
    return input


class AmplitudeEmbeddingBuilt(InputOperatorBuilt):
    """
    Will ignore the state of the circuit and set the features as the state.
    """

    def __init__(
        self,
        name: str,
        num_qubits: int,
        normalize: bool,
        pad_with: typing.Union[float, None],
        padding,
    ):
        super().__init__()
        self.num_qubits = num_qubits
        self.normalize = normalize
        self.pad_with = pad_with
        self.padding = padding
        self.named = True
        self.name = name

    def forward(self, _=None, **kwargs):
        x = kwargs[self.name]
        pad_len = 2**self.num_qubits - x.shape[-1]
        x = self.padding(
            input=x,
            pad=(0, pad_len),
            mode="constant",
            value=self.pad_with,  # type: ignore
        )
        x = self.normalize(x, p=2, dim=-1)  # type: ignore
        x = torch.complex(real=x, imag=torch.zeros_like(x))
        return x

    def __str__(self) -> str:
        return "AmplitudeEmbedding"

    def to_qasm(self) -> qasm.QasmRepresentation:
        raise NotImplementedError("AmplitudeEmbedding is not yet supported")


class AmplitudeEmbedding(InputOperator):
    """
    Amplitude Embedding
    """

    def __init__(
        self,
        name: str,
        qubits: list,
        normalize: bool = False,
        pad_with: typing.Union[float, None] = None,
    ):
        self.name = name
        self.named = True
        self.qubits = qubits
        self.pad_with = pad_with
        if pad_with is not None:
            self.padding = torch.nn.functional.pad
        else:
            self.padding = _identity
        if normalize:
            self.normalize = torch.nn.functional.normalize
        else:
            self.normalize = _identity

    def build(self, num_qubits: int) -> AmplitudeEmbeddingBuilt:
        assert num_qubits == len(
            self.qubits
        ), "Current Implementation requires all qubits to be used."
        return AmplitudeEmbeddingBuilt(
            num_qubits=num_qubits,
            normalize=self.normalize,
            pad_with=self.pad_with,
            padding=self.padding,
            name=self.name,
        )

    def __str__(self) -> str:
        return "AmplitudeEmbedding"

    def to_qasm(self) -> qasm.QasmRepresentation:
        raise NotImplementedError("AmplitudeEmbedding is not yet supported")


class AngleEmbeddingBuilt(InputOperatorBuilt):
    """
    Will ignore the state of the circuit and set the features as the state.
    """

    def __init__(self, name: str, num_qubits: int, qubits: typing.List[int], rotation: str):
        super().__init__()
        self.num_qubits = num_qubits
        self.qubits = qubits
        self.named = True
        self.name = name

        _0 = torch.tensor((0 + 0j,), requires_grad=False)
        _i = torch.tensor((0 + 1j,), requires_grad=False)
        self.register_buffer("_0", _0, persistent=False)
        self.register_buffer("_i", _i, persistent=False)
        self.rotation = rotation
        self.rots = utils.parse_rot(rotation)
        a, b, self.a_op, self.b_op = self._get_rot_matrices(self.rots, num_qubits, qubits)
        self.register_buffer("_a", a.contiguous(), persistent=False)
        self.register_buffer("_b", b.contiguous(), persistent=False)
        z = torch.zeros(2**num_qubits, dtype=torch.cfloat)
        z[0] = 1
        self.register_buffer("_z", z, persistent=False)

    @staticmethod
    def _get_rot_matrices(rot, num_qubits, qubits):
        gates = [rot(qubit=w, remapping=None).build(num_qubits) for w in qubits]
        a = [g._a.T for g in gates]
        b = [g._b.T for g in gates]
        a = torch.stack(a, dim=0)
        b = torch.stack(b, dim=0)
        return a, b, gates[0].a_op, gates[0].b_op

    def forward(self, state=None, **kwargs):
        inp = kwargs[self.name]
        a_matrix = torch.einsum("dab,...d->...dab", self._a, self.a_op(inp / 2))
        b_matrix = torch.einsum("dab,...d->...dab", self._b, self.b_op(inp / 2))
        matrix = a_matrix + b_matrix
        prod = self.prod(matrix)
        return prod @ self._z

    def prod(self, inp):
        """
        Return the matrix product along the third-to-last dimension.
        TODO: Optimize this.
        """
        res = inp[..., 0, :, :]
        for i in range(1, inp.shape[-3]):
            res = res @ inp[..., i, :, :]
        return res

    def to_qasm(self) -> qasm.QasmRepresentation:
        reps = []
        r = hash(random.random())
        for w in self.qubits:
            reps.append(self.rots(qubit=w, name=f"angle_{r}_{w}").to_qasm())
        return reps  # type: ignore

    def __str__(self) -> str:
        return f"AngleEmbedding_{self.rotation}"

    def decompose(self):
        reps = []
        r = hash(random.random())
        for w in self.qubits:
            reps.append(self.rots(qubit=w, name=f"angle_{r}_{w}"))
        return reps


class AngleEmbedding(InputOperator):
    """
    Angle Embedding
    """

    def __init__(
        self, name: str, qubits: typing.Union[typing.List[int], None] = None, rotation="rx"
    ):
        self.name = name
        self.named = True
        self.qubits = qubits
        self.rotation = rotation

    def build(self, num_qubits: int) -> AngleEmbeddingBuilt:
        if self.qubits is None:
            qubits = list(range(num_qubits))
        else:
            qubits = self.qubits
        return AngleEmbeddingBuilt(
            name=self.name, qubits=qubits, num_qubits=num_qubits, rotation=self.rotation
        )

    def to_qasm(self) -> qasm.QasmRepresentation:
        reps = []
        r = hash(random.random())
        for w in self.qubits:
            reps.append(self.rots(qubit=w, name=f"angle_{r}_{w}").to_qasm())
        return reps  # type: ignore

    def __str__(self) -> str:
        return f"AngleEmbedding_{self.rotation}"



import typing
import torch
import qandle.ansaetze.ansatz as ansatz
import qandle.operators as op
import qandle.utils as utils
import qandle.qasm as qasm

__all__ = ["SpecialUnitary"]


class SpecialUnitary(ansatz.UnbuiltAnsatz):
    """
    The hardware efficient SU(2) 2-local circuit as described in `the Qiskit docs <https://docs.quantum.ibm.com/api/qiskit/qiskit.circuit.library.EfficientSU2>`_.
    """

    def __init__(
        self,
        qubits: typing.Union[typing.List[int], None] = None,
        reps: int = 1,
        rotations=["ry"],
    ):
        super().__init__()
        self.qubits = qubits
        self.reps = reps
        self.rotations = [utils.parse_rot(r) for r in rotations]

    def build(self, *args, **kwargs) -> ansatz.BuiltAnsatz:
        return SpecialUnitaryBuilt(
            num_qubits=kwargs["num_qubits"],
            reps=self.reps,
            rotations=self.rotations,
            qubits=self.qubits,
        )

    def __str__(self) -> str:
        return "SU"

    def to_qasm(self) -> qasm.QasmRepresentation:
        return [g.to_qasm() for g in self.decompose()]  # type: ignore

    def decompose(self) -> typing.List[op.Operator]:
        if self.qubits is None:
            raise ValueError(
                "It is not specified on which qubits to apply the ansatz. Build the circuit to specify or pass them explicitly as a list"
            )
        outp = []
        for r in self.rotations:
            for w in self.qubits:
                outp.append(r(w))
        for _ in range(self.reps):
            for wi in range(len(self.qubits) - 1):
                outp.append(op.CNOT(self.qubits[wi], self.qubits[wi + 1]))
            for r in self.rotations:
                for w in self.qubits:
                    outp.append(r(w))
        return outp


class SpecialUnitaryBuilt(ansatz.BuiltAnsatz):
    def __init__(
        self,
        num_qubits: int,
        qubits: typing.Union[typing.List[int], None],
        reps: int,
        rotations: typing.List[op.Operator],
    ):
        super().__init__()
        self.num_qubits = num_qubits
        self.reps = reps
        self.rotations = rotations
        self.qubits = qubits or list(range(num_qubits))
        self.to_matrix, self.to_state = utils.get_matrix_transforms(num_qubits, self.qubits)
        layers = []
        for rep in range(reps + 1):
            block = []
            for w in self.qubits:
                for r in rotations:
                    block.append(r(w))
            layers.append(block)
        self.register_buffer(
            "cnots",
            self._get_cnot_matrix(qubits=self.qubits, num_qubits=num_qubits).contiguous(),
            persistent=False,
        )
        self.layers = torch.nn.ModuleList(
            [
                torch.nn.Sequential(*[la.build(num_qubits=num_qubits) for la in lay])
                for lay in layers
            ]
        )

    @staticmethod
    def _get_cnot_matrix(num_qubits: int, qubits: typing.List[int]):
        cnots = [
            op.BuiltCNOT._calculate_matrix(c=qubits[c], t=qubits[c + 1], num_qubits=num_qubits)
            for c in range(len(qubits) - 1)
        ]
        r = cnots[0]
        for c in cnots[1:]:
            r = torch.matmul(r, c)
        return r

    def forward(self, state: torch.Tensor) -> torch.Tensor:
        state = self.layers[0](state)
        for layer in self.layers[1:]:
            state = torch.matmul(self.cnots, state)
            state = layer(state)
        return state

    def __str__(self) -> str:
        return "SU"

    def to_qasm(self) -> qasm.QasmRepresentation:
        return [g.to_qasm() for g in self.decompose()]  # type: ignore

    def decompose(self) -> typing.List[op.Operator]:
        res = []
        res.extend(self.layers[0])
        for mod in self.layers[1:]:
            res.extend(mod)
            for w in range(len(self.qubits) - 1):
                res.append(op.CNOT(self.qubits[w], self.qubits[w + 1]))
        return res



import torch
import qandle
import pytest


def test_shapes():
    h, w = 17, 18
    for c_in in [1, 3, 10]:
        for c_out in [1, 3, 10, 15]:
            conv1 = qandle.QConv(in_channels=c_in, out_channels=c_out, padding=1)
            for batch_size in [1, 10]:
                inp = torch.rand(batch_size, c_in, h, w, dtype=torch.float)
                out = conv1(inp)
                assert out.shape == (batch_size, c_out, h, w)
                assert out.dtype == torch.float

            conv2 = qandle.QConv(in_channels=c_in, out_channels=c_out, padding=0)
            for batch_size in [1, 10]:
                inp = torch.rand(batch_size, c_in, h, w, dtype=torch.float)
                out = conv2(inp)
                assert out.shape == (batch_size, c_out, h - 2, w - 2)
                assert out.dtype == torch.float


def test_errors():
    conv = qandle.QConv(in_channels=3, out_channels=10, kernel_size=3, padding=1)
    with pytest.raises(ValueError):
        inp = torch.rand(10, 4, 17, 18)  # wrong number of input channels
        conv(inp)
    with pytest.raises(ValueError):
        inp = torch.rand(10, 3, 17)  # wrong number of dimensions 1
        conv(inp)
    with pytest.raises(ValueError):
        inp = torch.rand(10, 3, 17, 18, 19)  # wrong number of dimensions 2
        conv(inp)



import typing
import networkx as nx

import dataclasses

import qandle.splitter.grouping as grouping
import qandle.operators as op

cnot_types = typing.Union[op.CNOT, op.BuiltCNOT]


def split(
    circuit,
    max_qubits: int = 5,
) -> typing.Dict[int, "SubcircuitContainer"]:
    assert circuit.num_qubits > 1, "Can't split circuits with only 1 qubit"
    assert all(
        isinstance(layer, op.Operator) for layer in circuit.layers
    ), f"Unknown layer type in circuit: {[type(layer) for layer in circuit.layers if not isinstance(layer, op.Operator)]}"

    G = _construct_graph(circuit)
    G = grouping.groupnodes(G, max_qubits)
    assigned = _assign_to_subcircuits(G.nodes(), circuit.layers)
    normalized = _normalize_subcircuits(assigned)
    return normalized


def _construct_graph(circuit) -> nx.DiGraph:
    nodes = []
    circuit = circuit.decompose()
    for i, layer in enumerate(circuit.layers):
        if isinstance(layer, cnot_types):
            nodes.append(grouping.Node(layer.c, layer.t, i))

    _txt = (
        "No CNOTs in the circuit, instead of splitting, just use separate circuits for each qubit"
    )
    assert len(nodes) > 0, _txt
    used_qubits = set(n.c for n in nodes) | set(n.t for n in nodes)
    _txt = f"Circuit has qubits not used in CNOTs. Instead of splitting, just use separate circuits for separate qubits. Only used qubits are: {used_qubits}"
    assert len(used_qubits) == circuit.num_qubits, _txt
    G = nx.DiGraph()
    for n in nodes:
        G.add_node(n)  # All CNOTs are nodes

    nodes_per_qubit = {w: [] for w in range(circuit.num_qubits)}
    for n in nodes:
        nodes_per_qubit[n.c].append(n)
        nodes_per_qubit[n.t].append(n)

    for w in range(circuit.num_qubits):
        for i in range(len(nodes_per_qubit[w]) - 1):
            # CNOTs are dependent on the previous CNOT they share a qubit with
            G.add_edge(nodes_per_qubit[w][i], nodes_per_qubit[w][i + 1])
    return G


def _assign_to_subcircuits(
    nodes: typing.List[grouping.Node], original_circuit_layers: list
) -> typing.Dict[int, typing.List[op.Operator]]:
    """
    From the list of nodes (which have their new group assigned), sort the original layers into subcircuits
    """
    # create new subcircuits, one for each group
    subcircuits = {i: dict() for i in set(n.group for n in nodes)}
    node_per_index = {n.original_index: n for n in nodes}
    for i, layer in enumerate(original_circuit_layers):
        # insert CNOTS into the subcircuits
        if isinstance(layer, cnot_types):
            n = node_per_index[i]
            subcircuits[n.group][i] = layer

    def get_close_cnot(original_index) -> grouping.Node:
        w = original_circuit_layers[original_index].qubit
        previous_cnode_index, next_cnode_index = None, None
        for i, layer in enumerate(original_circuit_layers):
            if isinstance(layer, cnot_types) and (layer.c == w or layer.t == w):
                if i < original_index:
                    previous_cnode_index = i
                elif i > original_index:
                    next_cnode_index = i
                    break
        nearest_cnot_index = (
            previous_cnode_index if previous_cnode_index is not None else next_cnode_index
        )
        assert nearest_cnot_index is not None, f"Error. \n\
            original circuit {original_circuit_layers}\n\
            index {original_index},\
            previous_cnode_index {previous_cnode_index},\
            next_cnode_index {next_cnode_index},"
        return node_per_index[nearest_cnot_index]

    for i, layer in enumerate(original_circuit_layers):
        if not isinstance(layer, cnot_types):
            # insert non-CNOTs into the subcircuits
            close_n = get_close_cnot(i)
            subcircuits[close_n.group][i] = layer

    # flatten the subcircuits, removing empty dict entries
    flat_subc = dict()
    for subc in subcircuits.keys():
        flat_subc[subc] = []
        for i in sorted(subcircuits[subc].keys()):
            flat_subc[subc].append(subcircuits[subc][i])
    return flat_subc


@dataclasses.dataclass
class SubcircuitContainer:
    layers: list  # list of gates
    mapping: dict  # original index -> normalised index


def _normalize_subcircuits(assigned: dict):
    """
    subcircuits may have gaps in the qubit indices, remove those e.g.
    [CNOT(0,1), CNOT(4,5)] becomes [CNOT(0,1), CNOT(2,3)].
    Also returns a mapping from the original index to the normalised index
    (so {0:0, 1:1, 4:2, 5:3} in the example above)
    """

    def normalize_subcircuit(subcircuit: list) -> SubcircuitContainer:
        all_qubits = (
            set(g.c for g in subcircuit if isinstance(g, cnot_types))
            | set(g.t for g in subcircuit if isinstance(g, cnot_types))
            | set(g.qubit for g in subcircuit if not isinstance(g, cnot_types))
        )
        all_qubits = sorted(list(all_qubits))
        mapping = {q: i for i, q in enumerate(all_qubits)}
        new_layers = []
        for g in subcircuit:
            if isinstance(g, cnot_types):
                new_layers.append(op.CNOT(control=mapping[g.c], target=mapping[g.t]))
            elif isinstance(g, op.CustomGate):
                gnew = op.CustomGate(
                    qubit=mapping[g.qubit],
                    matrix=g.original_matrix,
                    num_qubits=len(all_qubits),
                    self_description=g.description,
                )
                new_layers.append(gnew)
            else:
                c = (
                    op.BUILT_CLASS_RELATION.T[type(g)]
                    if isinstance(g, op.BuiltOperator)
                    else type(g)
                )
                otherargs = dict()
                if hasattr(g, "theta") and g.theta is not None:
                    otherargs["theta"] = g.theta
                if hasattr(g, "name") and g.name is not None:
                    otherargs["name"] = g.name
                new_layers.append(c(qubit=mapping[g.qubit], **otherargs))
        return SubcircuitContainer(new_layers, mapping)

    normalized = {i: normalize_subcircuit(assigned[i]) for i in assigned.keys()}
    return normalized



import pennylane as qml
import torch
import qandle


def test_sel():
    """Test StronglyEntanglingLayers"""
    num_qubits = 4
    depth = 10
    pl_dev = qml.device("default.qubit.torch", wires=num_qubits)
    inp = torch.rand(2**num_qubits, dtype=torch.cfloat)
    inp = inp / inp.norm()
    weights = torch.rand(depth, num_qubits, 3)

    @qml.qnode(device=pl_dev, interface="torch")
    def pl_circuit():
        qml.QubitStateVector(inp, wires=range(num_qubits))
        qml.StronglyEntanglingLayers(weights=weights, wires=range(num_qubits))
        return qml.state()

    pl_result = pl_circuit().to(torch.cfloat)
    qandle_sel = qandle.StronglyEntanglingLayer(
        qubits=list(range(num_qubits)), depth=depth, q_params=weights, remapping=None
    ).build(num_qubits=num_qubits)
    qandle_result = qandle_sel(inp)
    assert torch.allclose(pl_result, qandle_result, rtol=1e-6, atol=1e-6)


def test_sel_to_matrix():
    num_qubits = 5
    depth = 11
    pl_dev = qml.device("default.qubit.torch", wires=num_qubits)
    weights = torch.rand(depth, num_qubits, 3)

    @qml.qnode(device=pl_dev, interface="torch")
    def pl_circuit():
        qml.StronglyEntanglingLayers(weights=weights, wires=range(num_qubits))
        return qml.state()

    pl_matrix = qml.matrix(qml.StronglyEntanglingLayers(weights=weights, wires=range(num_qubits)))
    qandle_sel = qandle.StronglyEntanglingLayer(
        qubits=list(range(num_qubits)), depth=depth, q_params=weights, remapping=None
    ).build(num_qubits=num_qubits)
    qandle_matrix = qandle_sel.to_matrix()
    assert torch.allclose(pl_matrix, qandle_matrix, rtol=1e-6, atol=1e-6)


def test_sel_sub():
    num_qubits = 5
    batch = 17

    qandle_sel1 = qandle.StronglyEntanglingLayer(qubits=[0, 1, 2, 3], depth=7).build(num_qubits=5)
    qandle_sel2 = qandle.StronglyEntanglingLayer(qubits=[1, 2, 3, 4], depth=7).build(num_qubits=5)
    qandle_sel3 = qandle.StronglyEntanglingLayer(qubits=[0, 1, 3, 4], depth=7).build(num_qubits=5)

    inp = torch.rand(batch, 2**num_qubits, dtype=torch.cfloat)
    inp = inp / torch.linalg.norm(inp, dim=1, keepdim=True)
    qandle_sel1(inp)
    qandle_sel2(inp)
    qandle_sel3(inp)


def test_sel_batched():
    num_qubits = 5
    depth = 7
    batch = 17
    pl_dev = qml.device("default.qubit.torch", wires=num_qubits)
    inp = torch.rand(batch, 2**num_qubits, dtype=torch.cfloat)
    inp = inp / torch.linalg.norm(inp, dim=1, keepdim=True)
    weights = torch.rand(depth, num_qubits, 3)

    @qml.qnode(device=pl_dev, interface="torch")
    def pl_circuit():
        qml.QubitStateVector(inp, wires=range(num_qubits))
        qml.StronglyEntanglingLayers(weights=weights, wires=range(num_qubits))
        return qml.state()

    pl_result = pl_circuit().to(torch.cfloat)
    qandle_sel = qandle.StronglyEntanglingLayer(
        qubits=list(range(num_qubits)), depth=depth, q_params=weights, remapping=None
    ).build(num_qubits=num_qubits)
    qandle_result = qandle_sel(inp)
    assert torch.allclose(pl_result, qandle_result, rtol=1e-6, atol=1e-6)


def test_sel_budget():
    num_qubits = 5
    depth = 4
    rots = ["rz", "ry", "rz"]
    sel = qandle.StronglyEntanglingLayer(
        qubits=list(range(num_qubits)), depth=depth, rotations=rots, num_qubits_total=num_qubits
    ).build(num_qubits=num_qubits)
    sel_budget = qandle.StronglyEntanglingLayerBudget(
        num_qubits_total=num_qubits,
        rotations=rots,
        param_budget=num_qubits * depth * len(rots),
        qubits=list(range(num_qubits)),
        control_gate_spacing=3,
    )
    sel_dec = sel.decompose()
    sel_budget_dec = sel_budget.layers
    assert len(sel_dec) == len(sel_budget_dec)
    ssel_dec = sorted(sel_dec, key=lambda x: str(x))
    ssel_budget_dec = sorted(sel_budget_dec, key=lambda x: str(x))
    for s, sb in zip(ssel_dec, ssel_budget_dec):
        assert type(s) is type(sb)
    inp = torch.rand(7, 2**num_qubits, dtype=torch.cfloat)
    inp = inp / torch.linalg.norm(inp, dim=1, keepdim=True)
    sel(inp)


def test_sel_general():
    num_qubits = 5
    qandle_sel_ub = qandle.StronglyEntanglingLayer(qubits=list(range(num_qubits)), depth=10)
    qandle_sel = qandle_sel_ub.build(num_qubits=num_qubits)
    inp = torch.rand(2**num_qubits, dtype=torch.cfloat)
    assert isinstance(qandle_sel(inp), torch.Tensor)
    assert isinstance(qandle_sel.decompose(), list)
    assert isinstance(qandle_sel.__str__(), str)


def test_twolocal():
    utwo = qandle.TwoLocal(qubits=list(range(4)))
    assert isinstance(utwo.decompose(), list)
    assert isinstance(utwo.__str__(), str)

    inp = torch.rand(2**4, dtype=torch.cfloat)
    inp = inp / inp.norm()
    two = utwo.build(num_qubits=4)
    assert isinstance(two(inp), torch.Tensor)
    assert isinstance(two.decompose(), list)
    assert isinstance(two.__str__(), str)
    two.to_qasm()

    inp2 = torch.rand(2**5, dtype=torch.cfloat)
    inp2 = inp2 / inp2.norm()
    two2 = utwo.build(num_qubits=5)
    assert isinstance(two2(inp2), torch.Tensor)


def test_su():
    for num_w in [2, 3, 6]:
        inp = torch.rand(2**num_w, dtype=torch.cfloat)
        inp = inp / inp.norm()
        for reps in [0, 1, 2, 10]:
            for rots in [["ry"], ["rz"], ["ry", "rx"]]:
                su = qandle.SU(reps=reps, rotations=rots).build(num_qubits=num_w)
                assert isinstance(su(inp), torch.Tensor)
                assert isinstance(su.decompose(), list)
                assert isinstance(su.__str__(), str)
                assert len(su.decompose()) == len(rots) * num_w * (1 + reps) + reps * (num_w - 1)
        for additional in [0, 1, 2]:
            su1 = qandle.SU(qubits=list(range(additional, num_w + additional)))
            inp = torch.rand(2 ** (num_w + additional * 2), dtype=torch.cfloat)
            inp = inp / inp.norm()
            su1.build(num_qubits=num_w + additional * 2)(inp)



import torch

from . import operators as op
from . import qasm

from .errors import UnbuiltGateError
import random

import warnings


class BuiltMeasurement(op.BuiltOperator):
    def __str__(self) -> str:
        return "M"

    def to_qasm(self) -> qasm.QasmRepresentation:
        f = "measure q[{}] -> c[{}]"
        i = hash(random.random())
        return [
            qasm.QasmRepresentation(gate_str=f.format(w, w), qasm3_outputs=f"measured_{i}_{w}")
            for w in range(self.num_qubits)
        ]  # type: ignore

    def decompose(self):
        warnings.warn(
            "Decomposed measurements a no-op. Consider acting on the resulting statevector directly",
            RuntimeWarning,
        )

        class SingleQubitMeasurementDummy(op.UnbuiltOperator, torch.nn.Module):
            def __init__(self, qubit: int):
                self.qubit = qubit
                super().__init__()

            def __str__(self) -> str:
                return "M"

            def to_qasm(self) -> qasm.QasmRepresentation:
                return qasm.QasmRepresentation(
                    gate_str=f"measure q[{self.qubit}] -> c[{self.qubit}]",
                    qasm3_outputs=f"measured_{hash(random.random())}_{self.qubit}",
                )

            def forward(self, state: torch.Tensor) -> torch.Tensor:
                warnings.warn(
                    "Forward on a decomposed measurement is a no-op. Consider acting on the resulting statevector directly"
                )
                return state

            def build(self, num_qubits: int, **kwargs):
                return self

        return [SingleQubitMeasurementDummy(w) for w in range(self.num_qubits)]

    def to_matrix(self):
        """
        No-op for measurements.
        """
        return torch.eye(2**self.num_qubits)


class UnbuiltMeasurement(op.UnbuiltOperator):
    def __str__(self) -> str:
        return "M"

    def to_qasm(self) -> str:
        raise UnbuiltGateError(f"Unbuilt {self.__class__.__name__} cannot be converted to qasm.")

    def build(self, num_qubits: int, **kwargs) -> BuiltMeasurement:
        return BUILT_CLASS_RELATION[self.__class__](num_qubits=num_qubits, **kwargs)  # type: ignore


class MeasureAllAbsolute(BuiltMeasurement):
    """
    Measure the absolute probability of the state-vector.
    """

    def forward(self, state: torch.Tensor) -> torch.Tensor:
        return state.abs().pow(2)


MeasureJointProbability = MeasureAllAbsolute  # alias


class MeasureState(BuiltMeasurement):
    """
    Measure the state-vector. As the state-vector is already a known, this measurement is a no-op.
    """

    def forward(self, state: torch.Tensor) -> torch.Tensor:
        return state


class MeasureProbability(UnbuiltMeasurement):
    r"""
    Measure the probability of each qubit being in the :math:`\ket{0}` state.
    Use :code:`1 - measure_probability` to get the probability of each qubit being in the :math:`\ket{1}` state.
    """


class MeasureProbabilityBuilt(BuiltMeasurement):
    def __init__(self, num_qubits: int):
        super().__init__()
        BuiltMeasurement.__init__(self)
        self.num_qubits = num_qubits
        self.qubits = list(range(num_qubits))
        indices0 = self._get_indices0()
        indices1 = ~indices0
        self.register_buffer("indices0", indices0.contiguous(), persistent=False)
        # currently unused
        self.register_buffer("indices1", indices1.contiguous(), persistent=False)

    def _get_indices0(self):
        indices = []
        for w in range(self.num_qubits):
            ind = (torch.arange(2**self.num_qubits) // (2**w)) % 2 == 0
            indices.append(ind)
        return torch.stack(indices, dim=0).flip(dims=(0,))

    def forward(self, state: torch.Tensor):
        rep = state.repeat(self.num_qubits, 1, 1)
        sums = rep * self.indices0[:, None, :]
        return (sums.abs().pow(2).sum(dim=-1).squeeze()).transpose(-1, 0)


BUILT_CLASS_RELATION = {MeasureProbability: MeasureProbabilityBuilt}



import torch
import typing
import qw_map
import copy


import qandle.config as config
import qandle.operators as op
import qandle.utils as utils
import qandle.qasm as qasm
from qandle.ansaetze.ansatz import UnbuiltAnsatz, BuiltAnsatz


__all__ = ["StronglyEntanglingLayer"]


class StronglyEntanglingLayer(UnbuiltAnsatz):
    """
    A strongly entangling layer, inspired by `this paper <https://arxiv.org/abs/1804.00633>`_.
    Consists of a series of single-qubit rotations on each qubit specified in the :code:`qubits` parameter, followed by a series of CNOT gates. A higher :code:`depth` will increase the expressivity of the circuit at the cost of more parameters and a linear increase in runtime.
    """

    def __init__(
        self,
        qubits: typing.List[int],
        num_qubits_total: typing.Union[int, None] = None,
        depth: int = 1,
        rotations=["rz", "ry", "rz"],
        q_params=None,
        remapping: typing.Union[typing.Callable, None] = config.DEFAULT_MAPPING,
    ):
        super().__init__()
        self.depth = depth
        self.qubits = qubits
        self.num_qubits = num_qubits_total
        self.rots = [utils.parse_rot(r) for r in rotations]
        if q_params is None:
            q_params = torch.rand(depth, len(qubits), len(rotations))
        self.q_params = q_params
        if remapping is None:
            remapping = qw_map.none
        self.remapping = remapping

    def build(self, *args, **kwargs) -> BuiltAnsatz:
        return StronglyEntanglingLayerBuilt(
            num_qubits=kwargs["num_qubits"],
            depth=self.depth,
            rotations=self.rots,
            q_params=self.q_params,
            remapping=self.remapping,
            qubits=self.qubits,
        )

    def __str__(self) -> str:
        return "SEL"

    def to_qasm(self) -> qasm.QasmRepresentation:
        raise NotImplementedError("TODO")

    def decompose(self) -> typing.List[op.UnbuiltOperator]:
        layers = []
        for d in range(self.depth):
            for wi, w in enumerate(self.qubits):
                for r in range(len(self.rots)):
                    layers.append(
                        self.rots[r](
                            qubit=w,
                            theta=self.q_params[d, wi, r],
                            remapping=self.remapping,
                        )
                    )
            layers.extend(
                StronglyEntanglingLayerBuilt._get_cnots(
                    self.qubits, self.num_qubits, d % (self.num_qubits - 1)
                )
            )
        return layers


class StronglyEntanglingLayerBuilt(BuiltAnsatz):
    def __init__(
        self,
        num_qubits: int,
        qubits: typing.List[int],
        depth: int,
        rotations: typing.List[op.UnbuiltParametrizedOperator],
        q_params: torch.Tensor,
        remapping: typing.Union[typing.Callable, None],
    ):
        super().__init__()
        self.num_qubits = num_qubits
        self.depth = depth
        self.rots = rotations
        self.qubits = qubits
        layers = []
        for d in range(depth):
            for wi, w in enumerate(self.qubits):
                for r in range(len(self.rots)):
                    layers.append(
                        self.rots[r](
                            qubit=w,
                            theta=q_params[d, wi, r],
                            remapping=remapping,  # type: ignore
                        ).build(num_qubits)
                    )
            layers.extend(self._get_cnots(self.qubits, self.num_qubits, d % (len(self.qubits) - 1)))
        self.mods = torch.nn.Sequential(*layers)

    def forward(self, state: torch.Tensor):
        return self.mods(state)

    @staticmethod
    def _get_cnots(
        qubits: typing.List[int], num_qubits_total: int, iteration: int
    ) -> typing.List[op.UnbuiltOperator]:
        assert iteration + 1 < num_qubits_total
        cnots = []
        for ci in range(len(qubits)):
            ti = (ci + iteration + 1) % len(qubits)
            cnots.append(op.CNOT(qubits[ci], qubits[ti]).build(num_qubits_total))
        return cnots

    def __str__(self) -> str:
        return "SEL"

    def to_qasm(self) -> str:
        return [g.to_qasm() for g in self.decompose()]  # type: ignore

    def decompose(self) -> typing.List[op.UnbuiltOperator]:
        layers = []
        for mod in self.mods:
            layers.append(copy.deepcopy(mod))
        return layers



import torch
from qandle import drawer
from qandle import splitter
from qandle import operators
from qandle import qasm
from qandle import utils
import warnings
import typing

__all__ = ["Circuit"]


class Circuit(torch.nn.Module):
    class DummyUnbuiltCircuit:
        """Helper class to decompose a circuit with subcircuits without building it."""

        def __init__(self, num_qubits: int, layers: typing.List[operators.Operator]):
            self.num_qubits = num_qubits
            self.layers = []
            for layer in layers:
                if hasattr(layer, "decompose"):
                    self.layers.extend(layer.decompose())  # type: ignore
                else:
                    self.layers.append(layer)

        def decompose(self):
            return self

    def __init__(
        self,
        layers: typing.List[operators.Operator],
        num_qubits=None,
        split_max_qubits=0,
        circuit=None,
    ):
        super().__init__()
        self.name = ""
        self.named = True
        if not hasattr(layers, "__iter__"):
            layers = [layers]
        if num_qubits is not None:
            self.num_qubits = num_qubits
        else:
            self.num_qubits = len(self._used_qubits(layers))
        if circuit is not None:
            self.circuit = circuit
        else:
            if split_max_qubits > 0 and self.num_qubits > split_max_qubits:
                self.circuit = SplittedCircuit(
                    num_qubits=self.num_qubits,
                    subcircuitdict=splitter.split(
                        self.DummyUnbuiltCircuit(self.num_qubits, layers),
                        max_qubits=split_max_qubits,
                    ),
                )
            else:
                self.circuit = UnsplittedCircuit(self.num_qubits, layers)

    def forward(self, state=None, **kwargs):
        return self.circuit.forward(state, **kwargs)

    def to_matrix(self, **kwargs):
        return self.circuit.to_matrix(**kwargs)

    def __matmul__(self, x):
        """Allows the circuit to be multiplied with a state tensor, directly calling forward. Not compatible with named inputs."""
        return self.circuit.forward(x)

    @staticmethod
    def _used_qubits(layers: typing.List[operators.Operator]) -> typing.Set[int]:
        import qandle.measurements

        qubits = set()
        for layer in layers:
            if isinstance(layer, operators.CNOT):
                qubits.add(layer.c)
                qubits.add(layer.t)
            elif isinstance(layer, operators.SWAP):
                qubits.add(layer.a)
                qubits.add(layer.b)
            elif isinstance(layer, operators.Controlled):
                qubits.add(layer.c)
                qubits.add(layer.t.qubits)
            elif hasattr(layer, "qubit"):
                qubits.add(layer.qubit)  # type: ignore
            elif hasattr(layer, "num_qubits"):
                qubits.update(range(layer.num_qubits))  # type: ignore
            elif hasattr(layer, "qubits"):
                qubits.update(layer.qubits)
            elif isinstance(
                layer,
                (
                    qandle.measurements.UnbuiltMeasurement,
                    qandle.measurements.BuiltMeasurement,
                ),
            ):
                pass  # ignore measurements, as they act on all qubits anyway
            else:
                raise ValueError(
                    f"Unknown layer type {type(layer)}, number of qubits could not be inferred. Pass :code:`num_qubits` to the circuit."
                )
        if len(qubits) == 0:
            raise ValueError(
                "Number of qubits could not be inferred from layers. Please provide num_qubits to the circuit directly."
            )
        return qubits

    def draw(self):
        """Shorthand for drawer.draw(self.circuit)"""
        return drawer.draw(self.circuit)

    def split(self, max_qubits):
        if isinstance(self.circuit, UnsplittedCircuit):
            splitdict = splitter.split(self.circuit.decompose(), max_qubits=max_qubits)
            return Circuit(
                layers=[],
                num_qubits=self.num_qubits,
                circuit=SplittedCircuit(self.num_qubits, splitdict),
            )
        else:
            warnings.warn("Circuit already split, returning original circuit.")
            return self

    def decompose(self):
        return Circuit(layers=[], num_qubits=self.num_qubits, circuit=self.circuit.decompose())

    def to_qasm(self):
        return self.circuit.to_qasm()

    def to_openqasm2(self) -> str:
        return qasm.convert_to_qasm(self, qasm_version=2, include_header=True)

    def to_openqasm3(self) -> str:
        return qasm.convert_to_qasm(self, qasm_version=3, include_header=True)


class UnsplittedCircuit(torch.nn.Module):
    """
    Main class for quantum circuits. This class is a torch.nn.Module, so it can be used like any other PyTorch module.
    """

    def __init__(self, num_qubits: int, layers: list):
        """
        Create a new quantum circuit, building all layers.
        """
        super().__init__()
        self.num_qubits = num_qubits
        self.state = torch.zeros(2**num_qubits, dtype=torch.complex64)
        self.state[0] = 1
        layers_built = self._build_layers(layers, num_qubits)
        self.layers = torch.nn.ModuleList(layers_built)

    @staticmethod
    def _build_layers(layers: list, num_qubits: int) -> typing.List[torch.nn.Module]:
        layers_built = []
        for la in layers:
            if hasattr(la, "build"):
                layers_built.append(la.build(num_qubits=num_qubits))
            else:
                layers_built.append(la)
        return layers_built

    def forward(self, state=None, **kwargs):
        """
        Run the circuit. If state is None, the initial state is used. Supply specific inputs to specific gates using the kwargs, of the form name:torch.Tensor
        """
        if state is None:
            state = self.state
        for mod in self.layers:
            if mod.named:
                state = mod(state, **kwargs)
            else:
                state = mod(state)
        return state

    def to_qasm(self) -> qasm.QasmRepresentation:
        reps = []
        for layer in self.layers:
            if hasattr(layer, "to_qasm"):
                reps.extend(layer.to_qasm())
        return reps  # type: ignore

    def decompose(self) -> "UnsplittedCircuit":
        """
        Decompose the circuit into a circuit without any subcircuits, returning a new, flat Circuit.
        """
        new_layers = []
        for layer in self.layers:
            if hasattr(layer, "decompose"):
                if isinstance(layer, (Circuit)):
                    new_layers.extend(layer.decompose().circuit.layers)
                else:
                    new_layers.extend(layer.decompose())
            else:
                new_layers.append(layer)
        return UnsplittedCircuit(self.num_qubits, new_layers)

    def split(self, **kwargs) -> typing.Union["SplittedCircuit", "UnsplittedCircuit"]:
        """
        Split the circuit into subcircuits, returning a CuttedCircuit if the circuit was split, or the original circuit if not.
        """
        if self.num_qubits == 1:
            warnings.warn("Can't split circuit with only 1 qubit")
            return self
        splitted = splitter.split(self, **kwargs)
        return SplittedCircuit(self.num_qubits, splitted)

    def to_matrix(self, **kwargs):
        """
        Get the matrix representation of the circuit.
        """
        return utils.reduce_dot(*[mod.to_matrix(**kwargs) for mod in self.layers])


class SplittedCircuit(UnsplittedCircuit):
    """
    Circuit with subcircuits, created by splitting a Circuit.
    """

    def __init__(
        self,
        num_qubits: int,
        subcircuitdict: typing.Dict[int, splitter.SubcircuitContainer],
    ):
        torch.nn.Module.__init__(self)
        self.num_qubits = num_qubits
        self.state = torch.zeros(2**num_qubits, dtype=torch.complex64)
        self.state[0] = 1

        self._check_qubit_order(subcircuitdict)

        self.subcircuits = torch.nn.ModuleList(
            [
                Subcircuit(num_qubits, subcircuitdict[i].mapping, subcircuitdict[i].layers)
                for i in subcircuitdict
            ]
        )

    @staticmethod
    def _check_qubit_order(subcircuitdict):
        for sc in subcircuitdict.values():
            keys_sorted = sorted(sc.mapping.keys())
            values = [sc.mapping[k] for k in keys_sorted]
            assert values == sorted(values), f"Mapping {sc.mapping} is not ordered."

    def forward(self, inp=None, **kwargs):
        if inp is None:
            inp = self.state
        for sc in self.subcircuits:
            inp = sc(inp, **kwargs)
        return inp

    def to_qasm(self) -> qasm.QasmRepresentation:
        reps = [sc.to_qasm() for sc in self.subcircuits]
        return reps  # type: ignore

    def to_matrix(self, **kwargs):
        raise NotImplementedError("Can't get matrix representation of a splitted circuit.")


class Subcircuit(torch.nn.Module):
    """
    Submodule for CuttedCircuit, contains a single subcircuits
    """

    def __init__(self, num_qubits_parent: int, mapping: typing.Dict[int, int], layers: list):
        """
        num_qubits_parent: number of qubits in the parent circuit
        mapping: mapping of qubits from parent circuit to subcircuit
        layers: layers of the subcircuit
        """
        super().__init__()
        self.num_qubits_parent = num_qubits_parent
        self.mapping = mapping
        layers_built = UnsplittedCircuit._build_layers(layers, num_qubits=len(mapping))
        self.to_matrix, self.to_state = utils.get_matrix_transforms(
            num_qubits_parent, list(mapping.keys())
        )

        self.layers = torch.nn.ModuleList(layers_built)

    def forward(self, x, **kwargs):
        batched = x.ndim == 2
        if not batched:
            x = x.unsqueeze(0)
        x = self.to_matrix(x)
        for mod in self.layers:
            if mod.named:
                x = mod(x, **kwargs)
            else:
                x = mod(x)
        x = self.to_state(x)
        if not batched:
            x = x.squeeze(0)
        return x

    def to_qasm(self) -> qasm.QasmRepresentation:
        import re

        reps = [mod.to_qasm() for mod in self.layers if hasattr(mod, "to_qasm")]
        for r in reps:
            if r.qubit not in ("", None):
                r.qubit = [k for k, v in self.mapping.items() if v == r.qubit][0]
            else:
                rstr = r.gate_str

                def new_qubit(x):
                    old_q = int(x.group(1))
                    new_q = [k for k, v in self.mapping.items() if v == old_q][0]
                    return f"q[{new_q}]"

                rstr = re.sub(r"q\[(\d+)\]", new_qubit, rstr)
                r.gate_str = rstr
        return reps  # type: ignore



class UnbuiltGateError(ValueError):
    """
    Raised when a gate is used in a circuit before it is built. Either build the gate manually using :code:`gate.build()` or use the gate in a circuit, which will build it automatically.
    """

    pass



import torch
import qandle
from qandle.test import check_qasm

op = qandle.operators


def get_circuit_cnot10():
    return qandle.Circuit(
        num_qubits=10,
        layers=[
            op.CNOT(0, 1),
            op.CNOT(0, 2),
            op.CNOT(1, 2),
            op.CNOT(2, 3),
            op.CNOT(3, 4),
            op.CNOT(4, 3),
            op.CNOT(3, 4),
            op.CNOT(4, 3),
            op.CNOT(3, 4),
            op.CNOT(4, 3),
            op.CNOT(3, 4),
            op.CNOT(4, 3),
            op.CNOT(3, 5),
            op.CNOT(5, 6),
            op.CNOT(6, 7),
            op.CNOT(6, 1),
            op.CNOT(7, 8),
            op.CNOT(8, 9),
        ],
    )


def test_nested_circuits():
    c1 = qandle.Circuit(
        layers=[
            qandle.RX(0),
            qandle.CNOT(0, 1),
        ],
        num_qubits=3,
    )
    c2 = qandle.Circuit(
        layers=[qandle.RY(1), qandle.StronglyEntanglingLayer(qubits=[1, 2])],
        num_qubits=3,
    )
    c1c2 = qandle.Circuit(layers=[c1, qandle.RZ(0), c2], num_qubits=3)
    split = c1c2.split(max_qubits=3)
    inp = torch.rand(2**3, dtype=torch.cfloat)
    inp = inp / torch.linalg.norm(inp, dim=-1, keepdim=True)
    res_c1c2 = c1c2(inp)
    res_split = split(inp)
    assert torch.allclose(res_c1c2, res_split, rtol=1e-6, atol=1e-6)


def test_splitter_1():
    orig_c = get_circuit_cnot10()
    split_c = orig_c.split(max_qubits=5)
    inp = torch.rand(2**orig_c.num_qubits, dtype=torch.cfloat)
    inp = inp / torch.linalg.norm(inp, dim=-1, keepdim=True)
    orig_res = orig_c(inp)
    split_res = split_c(inp)
    assert torch.allclose(orig_res, split_res, rtol=1e-6, atol=1e-6)
    assert torch.allclose(orig_res, orig_c @ inp, rtol=1e-6, atol=1e-6)
    assert torch.allclose(split_res, split_c @ inp, rtol=1e-6, atol=1e-6)
    check_qasm.check(orig_c)
    check_qasm.check(split_c)


def test_splitter_2():
    su = qandle.SU(reps=2, rotations=["rx", "ry"], qubits=list(range(10)))
    orig_c_big = qandle.Circuit(layers=[su])
    orig_c = orig_c_big.decompose()
    split_c = orig_c.split(max_qubits=5)
    orig_gates = list(orig_c.circuit.layers)
    split_gates = [sc.layers for sc in split_c.circuit.subcircuits]
    assert len(orig_gates) == sum([len(g) for g in split_gates])
    assert isinstance(split_c.draw(), str)
    check_qasm.check(orig_c)
    check_qasm.check(split_c)


def test_splitter_small():
    su = qandle.SU(qubits=list(range(3)), reps=2, rotations=["rx", "ry"])
    orig_c_big = qandle.Circuit(layers=su.decompose()).decompose()
    inp = torch.rand(2**orig_c_big.num_qubits, dtype=torch.cfloat)
    inp = inp / torch.linalg.norm(inp, dim=-1, keepdim=True)
    split_t2 = orig_c_big.split(max_qubits=2)
    split_t3 = orig_c_big.split(max_qubits=3)
    split_t4 = orig_c_big.split(max_qubits=4)

    out_orig = orig_c_big(inp)
    out_t2 = split_t2(inp)
    out_t3 = split_t3(inp)
    out_t4 = split_t4(inp)

    assert torch.allclose(out_orig, out_t2, rtol=1e-6, atol=1e-6)
    assert torch.allclose(out_orig, out_t3, rtol=1e-6, atol=1e-6)
    assert torch.allclose(out_orig, out_t4, rtol=1e-6, atol=1e-6)
    check_qasm.check(orig_c_big)
    check_qasm.check(split_t2)
    check_qasm.check(split_t3)


def test_circuit_dummy():
    c = qandle.Circuit(
        split_max_qubits=2,
        layers=[qandle.RX(0), qandle.CNOT(0, 2), qandle.RY(1), qandle.CNOT(1, 2)],
    )
    assert isinstance(c.circuit, qandle.qcircuit.SplittedCircuit)
    assert len(c.circuit.subcircuits) == 2



# ruff: noqa: F403 F401

from .stronglyentangling import StronglyEntanglingLayer
from .twolocal import TwoLocal
from .specialunitary import SpecialUnitary
from .stronglyentangling_budget import StronglyEntanglingLayerBudget

SU = SpecialUnitary  # alias



import qandle.operators as op
import abc
import typing
import qandle.utils as utils

__all__ = []


class Ansatz(op.Operator, abc.ABC):
    @abc.abstractmethod
    def decompose(self) -> typing.List[op.Operator]:
        pass


class UnbuiltAnsatz(Ansatz, op.UnbuiltOperator, abc.ABC):
    """An ansatz exposed to the user. Needs to be built before it can be used."""

    @abc.abstractmethod
    def build(self, num_qubits: int, **kwargs) -> "BuiltAnsatz":
        pass


class BuiltAnsatz(Ansatz, op.BuiltOperator, abc.ABC):
    """A built ansatz, ready to be used."""

    def to_matrix(self, **kwargs):
        decomposed = self.decompose()
        decomposed = [
            gate.build(num_qubits=self.num_qubits) if hasattr(gate, "build") else gate
            for gate in decomposed
        ]
        sub_matices = [gate.to_matrix(**kwargs) for gate in decomposed]
        return utils.reduce_dot(*sub_matices)
